#include  "buffer.h"
#include <vector>
#include  <cstring>
#include  <iostream>
#include  <fstream>
#include  <cctype>
#include  <cstdlib>
#include  <stdexcept>

namespace atlas
{ namespace interpreter
  {

String_pool::String_pool(size_t b)
: block_size(b),start(NULL),point(NULL),chars_left(0),prev(NULL)
{}

char* String_pool::store(const char* s,size_t len)
{ size_t n= len>=block_size ? len+1 : block_size;
            // size of new block if needed
  if (len>=chars_left) // then the current block cannot store |s|, so allocate
  { if (start!=NULL) // initial block needs no backing up
    { String_pool* p=new String_pool(block_size);
      p->prev=prev; p->start=start; // only these fields need backing up
      prev=p; // link to old blocks, needed for destruction
    }
    start=point=new char[n]; // now we can copy to where |point| points
    chars_left=n;
  }
  strncpy(point,s,len);  char* result=point;
  chars_left-=len+1;
  point+=len;  *point++='\0';
  return result;
}

String_pool::~String_pool(void)
{ delete prev; delete[] start; }

const Hash_table::id_type Hash_table::empty =
   ~static_cast<Hash_table::id_type>(0);

Hash_table::Hash_table(size_t init,size_t block_size)
: pool(block_size)
, mod(init<2?2:init), hash_tab(mod,empty), name_tab(0)
{ name_tab.reserve(max_fill());
}

Hash_table::id_type Hash_table::hash
        (const char* s, size_t l) const // |s| a string of length |l>0|
{ unsigned long r=static_cast<unsigned char>(*s++)%mod;
  while (l-->1) // one less, because we already read one character
    r=(static_cast<unsigned char>(*s++)+(r<<8))%mod;
  return static_cast<id_type>(r);
}

bool Hash_table::knows (const char* name) const
{ size_t l = std::strlen(name);
  id_type i,h=hash(name,l);
  while ((i=hash_tab[h])!=empty) // search breaks off at first empty slot
    if (std::strncmp(name_tab[i],name,l)==0 && name_tab[i][l]=='\0')
      return true; // identifier found in table
    else
      if (++h==mod) // move past occupied slot
        h=0; // possibly wrapping around
  return false;
}

Hash_table::id_type Hash_table::do_match
	(const char* str, size_t l, bool copy_string)
{ id_type i,h=hash(str,l);
  while ((i=hash_tab[h])!=empty)
    if (std::strncmp(name_tab[i],str,l)==0 && name_tab[i][l]=='\0')
      return i; /* identifier found in table */
    else if (++h==mod)
      h=0; /* move past occupied slot */
  if (name_tab.size()>=max_fill())
  { 
    { std::vector<id_type>(mod=2*mod-1,empty).swap(hash_tab); // make length odd
      for (size_t i=0; i<name_tab.size(); ++i)
      { int h=hash(name_tab[i],std::strlen(name_tab[i])); // rehash the string
        while (hash_tab[h]!=empty) 
          if (++h==mod)
            h=0; /* find empty slot */
        hash_tab[h]=i;
      }
      name_tab.reserve(max_fill());
    }
    h=hash(str,l); 
    while (hash_tab[h]!=empty) 
      if (++h==mod) h=0; /* find free slot */
  }
  hash_tab[h]= name_tab.size(); /* this number represents the new identifier */
  name_tab.push_back(copy_string ? pool.store(str,l) : str);
  return hash_tab[h];
}

BufferedInput::BufferedInput (std::istream& s)
:base_stream(s)
,line_buffer()
,p(NULL)
,prompt(NULL)
,prompt2(NULL)
,def_ext(NULL)
,temp_prompt(NULL)
,readline(NULL)
,add_hist(NULL)
,line_no(1)
,cur_lines(0)
,input_stack()
,input_files_seen()
,stream(&base_stream)
{}

BufferedInput::BufferedInput
(const char* pr, rl_type rl, add_hist_type ah,const char* pr2,const char*
de)
:base_stream(std::cin)
,line_buffer()
,p(NULL)
,prompt(pr)
,prompt2(pr2)
,def_ext(de)
,temp_prompt("")
,readline(rl)
,add_hist(ah)
,line_no(1)
,cur_lines(0)
,input_stack()
,input_files_seen()
,stream(&base_stream)
{}

BufferedInput::~BufferedInput()
{ while (not input_stack.empty())
  { delete input_stack.back().stream;
    input_stack.pop_back();
  }
}

void BufferedInput::pop_file()
{ line_no = input_stack.back().line_no;
  std::cout << "Completely read file '" << input_stack.back().name
            << "'." << std::endl;
  input_files_seen.match(input_stack.back().name.c_str(),
                         input_stack.back().name.size()); // reading succeeded
  delete input_stack.back().stream;
  input_stack.pop_back();
  stream= input_stack.empty() ? &base_stream : input_stack.back().stream;
}

void BufferedInput::close_includes()
{ while (not input_stack.empty())
  { std::cerr << "Abandoning reading of file '" << input_stack.back().name
              << "' at line " << line_no << std::endl;
    line_no = input_stack.back().line_no;
    delete input_stack.back().stream;
    input_stack.pop_back();
  }
  stream= &base_stream;
}

input_record::input_record
(const char* file_name, const char* def_ext,unsigned long line) :
 name(file_name), stream (new std::ifstream(file_name)), line_no(line)
{ try
  {
    if (def_ext!=NULL and not stream->is_open())
    { long inx=name.size()-std::strlen(def_ext);
      if (inx<0 or name.substr(inx)!=def_ext) // only add extension if absent
      { name += def_ext;
        stream->open(name.c_str());
      }
    }
  }
  catch (...)
  { delete stream; throw; }
}

bool BufferedInput::push_file(const char* name, bool skip_seen)
{
  if (skip_seen and input_files_seen.knows(name))
    return true;
  input_stack.push_back(input_record(name,def_ext,line_no+cur_lines));
  // record where reading will resume
  if (input_stack.back().stream->good())
  { const std::string& name_ext=input_stack.back().name; bool avoid=false;
    for (unsigned i=input_stack.size()-1; i-->0; )
      if (name_ext==input_stack[i].name)
        avoid=true; // avoid recursive inclusion of active file
    if (avoid or
        (skip_seen and input_files_seen.knows(name_ext.c_str()))) // seen before
    { delete input_stack.back().stream;
      input_stack.pop_back();
    } // close file and pop record
    else
    { std::cout << "Starting to read from file '" << input_stack.back().name
             << "'."  << std::endl;
    stream= input_stack.back().stream;
      line_no=1; // prepare to read from pushed file
      cur_lines=0;
        // so we won't advance |line_no| when getting first line of new file
    }
    return true; // succeed whether or not a file was actually pushed
  }
  else
  { std::cerr << "failed to open input file '" << input_stack.back().name
              << "'." << std::endl;
    delete input_stack.back().stream;
    input_stack.pop_back();
// no need to call |pop_file|: |stream|, |line_no| and |cur_lines| unchanged
    return false;
  }
}

bool BufferedInput::getline()
{ line_buffer="";
  line_no+=cur_lines; cur_lines=0;
  const char* pr=
               (temp_prompt.empty() ? prompt : (temp_prompt+" "+prompt2).c_str())
;
  bool go_on, popped=false;
  do
  { std::string line;
  
  { if (stream->good())
      { 
        if (input_stack.empty() and prompt!=NULL)
          // do only at top level, and only if prompt enabled
        { prompt_length= std::strlen(pr);
          if (readline!=NULL) // skip calling 'readline' if no function is supplied
          { char* l=readline(pr);
            if (l==NULL) // then |readline| failed, flag end of file
            { line=""; stream->setstate(std::ios_base::eofbit);
              std::cout << "^D\n";
            }
            else
            { line=l;
              if (add_hist!=NULL and *l!='\0')
                add_hist(l);
              else
                std::free(l);
            }
          }
          else
          { std::cout << pr;
              std::getline(std::cin,line,'\n'); }
        }
        else
           std::getline(*stream,line,'\n');
   }
    if (line.empty() and not stream->good()) // nothing found and file ended
    { if (cur_lines>0)
      { std::cerr << "end of file after backslash ";
        if (input_stack.empty())
          std::cerr << (stream==&std::cin ? "on standard input" : "in main file")
                    << std::endl;
        else
        { std::cerr << "in file '" << input_stack.back().name << '\'' << std::endl;
        pop_file(); popped=true;
        }
      }
      else if (input_stack.empty())
        return false; // all hope of returning something has gone
      else { pop_file(); popped=true; }
      break;
  // in these cases terminate line-continuation; return the part that was found
    }
    else {} // something (at least a bona fide empty line) was contributed here
  }
    ++cur_lines;
    std::string::size_type l=line.length();
    while (l>0 and std::isspace(line[l-1]))
      --l;
    if (l<line.length())
      line.erase(l); // remove trailing space
    pr= "\\ "; // continuation prompt
    if (go_on=(l>0 and line[l-1]=='\\'))
      // keep going only if final backslash present
      line.erase(l-1); // in which case it is chopped off
    line_buffer+=line; // append |line| to buffer in all cases
  }
  while(go_on);
  line_buffer.push_back(popped ? '\f' :'\n');
     // add suppressed newline, or form feed if file was popped
  p=line_buffer.data();
return true;
  // delay reporting end of input if anything was read at all
}

void BufferedInput::locate (const char* p, int& line, int& column) const
{ line=line_no; column=p-line_buffer.data(); }

void BufferedInput::show_range
  (std::ostream& out,unsigned long l0, int c0, unsigned long l1, int c1)
  const
{ if (l1==line_no)
  { int pl=prompt_length; // offset of last line on the screen
    if (stream!=&std::cin or cur_lines>1)
    { if (not input_stack.empty())
        out << "In input file '" << input_stack.back().name << "', ";
      if (stream!=&std::cin)
        out << "line " << l0 << ":\n";
      out<<line_buffer; pl=0; // echo line in these cases
    }
    if (l0<l1)
      c0=0; // forget start column of multi-line range
    for (int i=pl+c0; i>0; --i)
      out<<' ';
    for (int i=c1; i>c0; --i)
      out<<'^';
    out << std::endl;
    if (l0<l1)
      if (l1-l0==1)
        out << "Range started in previous line\n";
      else
        out << "Range started " << (l1-l0) << "lines above\n";
  }
  else // range ended before the current line
    out << "Range from line " << l0 << " column " << c0 
	<< " to line " << l1 << " column " << c1 << ".\n";
}

void BufferedInput::push_prompt(char c) 
{ temp_prompt.push_back(c); }
char BufferedInput::top_prompt() const
{ size_t l=temp_prompt.length();
  return l==0 ? '\0' : temp_prompt[l-1];
}
void BufferedInput::pop_prompt()
{ size_t l=temp_prompt.length();
  if (l>0)
    temp_prompt.erase(l-1);
}
void BufferedInput::reset() 
{ temp_prompt=""; }


BufferedInput* main_input_buffer=NULL;

  }
}

