#include  "evaluator.h"
#include  <cstdlib>
#include  <algorithm>
#include  "built-in-types.h"

namespace atlas { namespace interpreter {

layer::list layer::lexical_context;

type_expr last_type;
shared_value last_value;

namespace {

class frame
{
  const id_pat& pattern;
public:
  static shared_context current;

  frame (const id_pat& pattern)
  : pattern(pattern)
  { current = std::make_shared<evaluation_context>(std::move(current)); }
  ~frame() { current = current->tail(); } // don't move here!

  void bind (const shared_value& val)
    { current->reserve(count_identifiers(pattern));
      thread_components(pattern,val,current->back_inserter());
    }
};

class lambda_frame
{
  const id_pat& pattern;
  const shared_context saved;
public:
  lambda_frame (const id_pat& pattern, const shared_context& outer)
  : pattern(pattern), saved(std::move(frame::current))
  { assert(&outer!=&frame::current); // for excluded case use |frame| instead
    try
    { frame::current = std::make_shared<evaluation_context>(outer); }
    catch(...)
    { frame::current = std::move(saved); throw; }
    // restore as destructor would do
  }
  ~lambda_frame() { frame::current = std::move(saved); }

  void bind (const shared_value& val)
    { frame::current->reserve(count_identifiers(pattern));
      thread_components(pattern,val,frame::current->back_inserter());
    }
};


shared_context frame::current; 


id_type equals_name()
{ static id_type name=main_hash_table->match_literal("=");
  return name;
}

id_type size_of_name()
{ static id_type name=main_hash_table->match_literal("#");
  return name;
}
id_type print_name()
{ static id_type name=main_hash_table->match_literal("print");
  return name;
}
id_type prints_name()
{ static id_type name=main_hash_table->match_literal("prints");
  return name;
}

inline bool is_special_operator(id_type id)
{ return id==size_of_name()
        or id==print_name()
        or id==prints_name(); }

bool can_coerce_arg
  (expression e,size_t i,const type_expr& from,type_expr& to)
{ if (to.specialise(from))
    return true; // type matches without coercion
  tuple_expression* tup= dynamic_cast<tuple_expression*>(e);
  if (tup==nullptr or tup->component.size()!=2)
    return false; // we need a pair to insert a coercion
  return coerce(from,to,tup->component[i]);
}

void print_wrapper(expression_base::level l)
{
  *output_stream << *execution_stack.back() << std::endl;
  if (l!=expression_base::single_value)
    push_expanded(l,pop_value()); // remove and possibly expand value
}

void prints_wrapper(expression_base::level l)
{ shared_value v=pop_value();
  const string_value* s=dynamic_cast<const string_value*>(v.get());
  if (s!=nullptr)
    *output_stream << s->val << std::endl;
  else
  { const tuple_value* t=dynamic_cast<const tuple_value*>(v.get());
    if (t!=nullptr)
    { for (auto it=t->val.begin(); it!=t->val.end(); ++it)
      { s=dynamic_cast<const string_value*>(it->get());
        if (s!=nullptr)
	  *output_stream << s->val;
        else
           *output_stream << *it->get();
      }
      *output_stream << std::endl;
    }
    else
      *output_stream << *v << std::endl; // just like |print| in other cases
  }
  if (l==expression_base::single_value)
    wrap_tuple<0>(); // don't forget to return a value if asked for
}

void sizeof_wrapper(expression_base::level l)
{ size_t s=get<row_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(s));
}

void suffix_element_wrapper(expression_base::level l)
{ shared_value e=pop_value();
  own_row r=get_own<row_value>();
  if (l!=expression_base::no_value)
  { r->val.push_back(e);
    push_value(r);
  }
}

void prefix_element_wrapper(expression_base::level l)
{ own_row r=get_own<row_value>();
  shared_value e=pop_value();
  if (l!=expression_base::no_value)
  { r->val.insert(r->val.begin(),e);
    push_value(r);
  }
}

void join_rows_wrapper(expression_base::level l)
{ shared_row y=get<row_value>();
  shared_row x=get<row_value>();
  if (l!=expression_base::no_value)
  { own_row result = std::make_shared<row_value>(x->val.size()+y->val.size());
    std::copy(y->val.begin(),y->val.end(), 
     std::copy(x->val.begin(),x->val.end(),result->val.begin()));
  push_value(std::move(result));
  }

}

id_pat copy_id_pat(const id_pat& p)
{
  containers::sl_list<id_pat> rl;
  std::transform(p.sublist.begin(),end(p.sublist)
                , std::back_inserter(rl), copy_id_pat);
  return id_pat (p.name, p.kind, rl.undress());
}

inline bool functype_specialise
  (type_expr& t, const type_expr& from, const type_expr& to)
{ return t.specialise(gen_func_type) 
  and t.func->arg_type.specialise(from) 
  and t.func->result_type.specialise(to);
}

}

void reset_evaluator()
{ if (!execution_stack.empty())
  { if (verbosity==0)
      execution_stack.clear();
    else
    {
      std::cerr << "Discarding from execution stack:" << std::endl;
      do
        std::cerr << *pop_value() << std::endl;
      while (!execution_stack.empty());
    }
  }
}

void layer::add(id_type id,type_expr&& t)
{ for (auto it=variable.begin(); it!=variable.end(); ++it)
  // traverse |variable| vector
    if (it->first==id)
      throw program_error 
       (std::string("Multiple binding of ")
                    +main_hash_table->name_of(id)
                    +" in same scope");
  variable.emplace_back( id, std::move(t) );
}

const_type_p layer::lookup
  (id_type id, size_t& depth, size_t& offset)
{ size_t i=0;
  for (auto range=lexical_context.cbegin(); not lexical_context.at_end(range);
       ++range,++i)
    for (auto it=(*range)->cbegin(); it!=(*range)->cend(); ++it)
      if (it->first==id)
      { depth=i;
        offset=it-(*range)->begin();
        return &it->second;
      }
  return nullptr;
}

void layer::specialise (size_t depth, size_t offset,const type_expr& t)
{ auto range=lexical_context.cbegin();
  while (depth-->0)
    ++range;
  (**range)[offset].second.specialise(t);
}

expression_ptr convert_expr(const expr& e, type_expr& type)
{
  switch(e.kind)
  {
   
   case integer_denotation:
     { expression_ptr d(new denotation
         (std::make_shared<int_value>(e.int_denotation_variant)));
       return conform_types(int_type,type,std::move(d),e);
     }
   case string_denotation:
     { expression_ptr d(new denotation
         (std::make_shared<string_value>(e.str_denotation_variant)));
       return conform_types(str_type,type,std::move(d),e);
     }
   case boolean_denotation:
     { expression_ptr d(new denotation
           (std::make_shared<bool_value>(e.int_denotation_variant)));
       return conform_types(bool_type,type,std::move(d),e);
     }
   
   case last_value_computed:
       return conform_types(last_type,type
                           ,expression_ptr(new denotation(last_value))
                           ,e);
   
   case tuple_display:
   { type_expr tup=unknown_tuple(length(e.sublist));
     tuple_expression* tup_exp;
     expression_ptr result(tup_exp=new tuple_expression(0));
     std::vector<expression_ptr>& comp = tup_exp->component;
     comp.reserve(length(e.sublist));
     bool tuple_expected = type.specialise(tup);
     // whether |type| is a tuple of correct size
     type_list::iterator tl_it =
       (tuple_expected ? &type : &tup)->tupple.begin();
     for (auto it=e.sublist.begin(); not e.sublist.at_end(it); ++it,++tl_it)
       comp.push_back(convert_expr(*it,*tl_it));
     if (tuple_expected or coerce(tup,type,result))
         return result;
     throw type_error(e,tup.copy(),std::move(type));
   }
   
   case list_display:
   { std::unique_ptr<list_expression> result (new list_expression(0));
     result->component.reserve(length(e.sublist));
     if (type==void_type)
     { type_expr target; // initially undetermined common component type
       for (auto it=e.sublist.begin(); not e.sublist.at_end(it); ++it)
         result->component.push_back(convert_expr(*it,target));
       return expression_ptr(new voiding(std::move(result)));
       // and forget |target|
     }
     else if (type.specialise(row_of_type))
     { for (auto it=e.sublist.begin(); not e.sublist.at_end(it); ++it)
         result->component.push_back(convert_expr(*it,*type.component_type));
       return std::move(result);
     }
     
     { type_expr comp_type;
       const conversion_record* conv = row_coercion(type,comp_type);
       if (conv==nullptr)
       // no conversion was found; there's nothing left but to report an error
         throw type_error(e,row_of_type.copy(),std::move(type));
     
       for (auto it=e.sublist.begin(); not e.sublist.at_end(it); ++it)
         result->component.push_back(convert_expr(*it,comp_type));
       return expression_ptr(new
         conversion(*conv,expression_ptr(std::move(result))));
     }
   }
   
   case applied_identifier:
   { const id_type id=e.identifier_variant;
     const_type_p id_t; size_t i,j;
     const bool is_local=(id_t=layer::lookup(id,i,j))!=nullptr;
     if (not is_local and (id_t=global_id_table->type_of(id))==nullptr)
       throw program_error  
          (std::string("Undefined identifier ")
   	+main_hash_table->name_of(id));
   
     expression_ptr id_expr = is_local
     ? expression_ptr(new local_identifier(id,i,j))
   : expression_ptr(new global_identifier(id));
     if (type.specialise(*id_t)) // then required type admits known identifier type
       { if (type!=*id_t)
         // usage has made type of identifier more specialised
         { if (is_local)
             layer::specialise(i,j,type);
             // then refine its type in the local context
           else
             global_id_table->specialise(id,type); // or in |global_id_table|
         }
         return id_expr;
       }
     else if (coerce(*id_t,type,id_expr))
       return id_expr;
     throw type_error(e,id_t->copy(),std::move(type));
   }
   
   case function_call:
   { if (e.call_variant->fun.kind==applied_identifier)
       
       { const id_type id =e.call_variant->fun.identifier_variant;
         size_t i,j; // dummies; local binding not used here
         if (layer::lookup(id,i,j)==nullptr) // not calling by local identifier
         { const overload_table::variant_list& variants
             = global_overload_table->variants(id);
           if (variants.size()>0 or is_special_operator(id))
             return expression_ptr(resolve_overload(e,type,variants));
         }
       }
     type_expr f_type=gen_func_type.copy(); // start with generic function type
     expression_ptr fun = convert_expr(e.call_variant->fun,f_type);
     expression_ptr arg = convert_expr(e.call_variant->arg,f_type.func->arg_type);
     expression_ptr call (new call_expression(std::move(fun),std::move(arg)));
     return conform_types(f_type.func->result_type,type,std::move(call),e);
   }
   
   case let_expr:
   { const let& lexp=e.let_variant;
     id_pat& pat=lexp->pattern;
     type_expr decl_type=pattern_type(pat);
     expression_ptr arg = convert_expr(lexp->val,decl_type);
   layer new_layer(count_identifiers(pat));
     thread_bindings(pat,decl_type,new_layer);
     return expression_ptr(new 
       let_expression(pat,std::move(arg),convert_expr(lexp->body,type)));
   }
   
   case lambda_expr:
   { const lambda& fun=e.lambda_variant;
     const id_pat& pat=fun->pattern;
     const type_expr& arg_type=fun->parameter_type;
   layer new_layer(count_identifiers(pat));
     thread_bindings(pat,arg_type,new_layer);
     if (type!=void_type)
     { if (not (type.specialise(gen_func_type)
                and type.func->arg_type.specialise(arg_type)))
       throw type_error(e,
                          type_expr(arg_type.copy(),unknown_type.copy()),
                          std::move(type));
       return expression_ptr(new 
         lambda_expression(pat, convert_expr(fun->body,type.func->result_type)));
     }
     else
     { type_expr dummy; // unused result type
       expression_ptr result(new 
         lambda_expression(pat, convert_expr(fun->body,dummy)));
       return expression_ptr(new voiding(std::move(result)));
     }
   }
   
   case seq_expr:
   { const sequence& seq=e.sequence_variant;
     if (seq->forward)
     { expression_ptr first = convert_expr(seq->first,as_lvalue(void_type.copy()));
       expression_ptr last  = convert_expr(seq->last,type);
       return expression_ptr(new
         seq_expression(std::move(first),std::move(last)));
     }
     else
     { expression_ptr first = convert_expr(seq->first,type);
       expression_ptr last  = convert_expr(seq->last,as_lvalue(void_type.copy()));
       return expression_ptr(new
          next_expression(std::move(first),std::move(last)));
     }
   }
   
   case subscription:
   { type_expr array_type, index_type, subscr_type;
       // all initialised to |undetermined_type|
     expression_ptr array
       = convert_expr(e.subscription_variant->array,array_type);
     expression_ptr index
       = convert_expr(e.subscription_variant->index,index_type);
     subscr_base::sub_type kind;
     expression_ptr subscr;
     if (subscr_base::indexable(array_type,index_type,subscr_type,kind))
       switch (kind)
       { case subscr_base::row_entry:
         subscr.reset(new row_subscription(std::move(array),std::move(index)));
       break;
       case subscr_base::vector_entry:
         subscr.reset(new vector_subscription(std::move(array),std::move(index)));
       break;
       case subscr_base::ratvec_entry:
         subscr.reset(new ratvec_subscription(std::move(array),std::move(index)));
       break;
       case subscr_base::string_char:
         subscr.reset(new string_subscription(std::move(array),std::move(index)));
       break;
       case subscr_base::matrix_entry:
         subscr.reset(new matrix_subscription(std::move(array),std::move(index)));
       break;
       case subscr_base::matrix_column:
         subscr.reset(new matrix_slice(std::move(array),std::move(index)));
       break;
       case subscr_base::mod_poly_term:
         subscr.reset(new module_coefficient(std::move(array),std::move(index)));
       break;
       }
     else
     { std::ostringstream o;
       o << "Cannot subscript value of type " << array_type 
         << " with index of type " << index_type;
       throw expr_error(e,o.str());
     }
   
     return conform_types(subscr_type,type,std::move(subscr),e);
   }
   
   case conditional_expr:
   { expression_ptr c  =
       convert_expr(e.if_variant->condition,as_lvalue(bool_type.copy()));
     type_expr else_type(type.copy());
     // make a copy so as to treat branches similarly
     expression_ptr th = convert_expr(e.if_variant->then_branch,type);
     expression_ptr el = convert_expr(e.if_variant->else_branch,else_type);
     if (type!=else_type) // we had different specialisations in the two branches
     {
       if (type==void_type or else_type==void_type)
         // type was unknown, one branch became void
       {
         if (type==void_type)
           el.reset(new voiding(std::move(el)));
         else
           th.reset(new voiding(std::move(th))),type = std::move(else_type);
       }
       else if (type==unknown_type or else_type==unknown_type)
         // error exit is always OK
         type.specialise(else_type); // but retain the more specific one in |type|
       else
       {
         int cmp = is_close(type,else_type);
         if ((cmp&0x1)!=0) // \.{then} branch may convert to |else_type|
           th =
             convert_expr(e.if_variant->then_branch,type = std::move(else_type));
         else if ((cmp&0x2)!=0) // |else| branch may convert to |type|
           el = convert_expr(e.if_variant->else_branch,type);
         else
         { std::ostringstream o;
           o << "Could not find common type for branches of conditional, "
           "types are " << type << " and " << else_type;
           throw expr_error(e,o.str());
         }
       }
     }
     return expression_ptr(new
       conditional_expression(std::move(c),std::move(th),std::move(el)));
   }
   
   case while_expr:
   { const w_loop& w=e.while_variant;
     expression_ptr c = convert_expr(w->condition,as_lvalue(bool_type.copy()));
     if (type==void_type)
     { expression_ptr result(new 
          while_expression(std::move(c),convert_expr(w->body, as_lvalue(void_type.copy()))));
       return expression_ptr(new voiding(std::move(result)));
     }
     else if (type.specialise(row_of_type))
     { expression_ptr b = convert_expr(w->body, *type.component_type);
       return expression_ptr (new while_expression(std::move(c),std::move(b)));
     }
     else
     
     { type_expr comp_type;
       const conversion_record* conv = row_coercion(type,comp_type);
       if (conv==nullptr)
         throw type_error(e,row_of_type.copy(),std::move(type));
     
       return expression_ptr(new conversion(*conv, expression_ptr
         (new while_expression(std::move(c),convert_expr(w->body,comp_type)))));
     }
   }
   
   case for_expr:
   { const f_loop& f=e.for_variant;
     type_expr in_type;
     expression_ptr in_expr = convert_expr(f->in_part,in_type);  // \&{in} part
     subscr_base::sub_type which; // the kind of aggregate iterated over
     layer bind(count_identifiers(f->id));
      // for identifier(s) introduced in this loop
     
     { type_expr comp_type; const type_expr* tp;
       if (subscr_base::indexable(in_type,*(tp=&int_type),comp_type,which) 
        or subscr_base::indexable(in_type,*(tp=&param_type),comp_type,which))
       { type_expr pt = pattern_type(f->id);
         type_list it_comps;
         it_comps.push_front(std::move(comp_type));
         it_comps.push_front(type_expr(tp->copy()));
         type_expr it_type(std::move(it_comps));
         if (not pt.specialise(it_type))
           throw expr_error(e,"Improper structure of loop variable pattern");
         thread_bindings(f->id,it_type,bind);
       }
       else
       { std::ostringstream o;
         o << "Cannot iterate over value of type " << in_type;
         throw expr_error(e,o.str());
       }
     }
     type_expr body_type;
     type_expr *btp=&body_type; // point to place to record body type
     const conversion_record* conv=nullptr;
     if (type==void_type)
       btp=&type; // we can reuse this type; no risk of specialisation
     else if (type.specialise(row_of_type))
       btp=type.component_type;
     else if ((conv=row_coercion(type,body_type))==nullptr)
       throw type_error(e,row_of_type.copy(),std::move(type));
     expression_ptr body(convert_expr (f->body,*btp));
   expression_ptr loop(new
       for_expression(f->id,std::move(in_expr),std::move(body),which));
   return type==void_type ? expression_ptr(new voiding(std::move(loop)))
     :  conv!=nullptr ? expression_ptr(new conversion(*conv,std::move(loop)))
     :  std::move(loop) ;
   }
   
   case cfor_expr:
   { const c_loop& c=e.cfor_variant;
     expression_ptr count_expr = convert_expr(c->count,as_lvalue(int_type.copy()));
     static const shared_value zero = std::make_shared<int_value>(0);
       // avoid repeated allocation
     expression_ptr bound_expr = is_empty(c->bound) 
       ? expression_ptr(new denotation(zero))
       : convert_expr(c->bound,as_lvalue(int_type.copy())) ;
   
     layer bind(1); bind.add(c->id,int_type.copy());
     type_expr body_type;
     type_expr *btp=&body_type; // point to place to record body type
     const conversion_record* conv=nullptr;
     if (type==void_type)
       btp=&type; // we can reuse this type; no risk of specialisation
     else if (type.specialise(row_of_type))
       btp=type.component_type;
     else if ((conv=row_coercion(type,body_type))==nullptr)
       throw type_error(e,row_of_type.copy(),std::move(type));
     expression_ptr body(convert_expr (c->body,*btp));
   expression_ptr loop;
     if (c->up)
       loop.reset(new  inc_for_expression
         (c->id,std::move(count_expr),std::move(bound_expr),std::move(body)));
     else
       loop.reset(new  dec_for_expression
         (c->id,std::move(count_expr),std::move(bound_expr),std::move(body)));
     return type==void_type ? expression_ptr(new voiding(std::move(loop))) : 
            conv!=nullptr ? expression_ptr(new conversion(*conv,std::move(loop)))
                          :  std::move(loop);
   
   }
   
   case cast_expr:
   { const cast& c=e.cast_variant;
     expression_ptr p = convert_expr(c->exp,c->type);
     return conform_types(c->type,type,std::move(p),e);
   }
   
   case op_cast_expr:
   { const op_cast& c=e.op_cast_variant;
     const overload_table::variant_list& variants =
      global_overload_table->variants(c->oper);
     type_expr& ctype=c->type;
     if (is_special_operator(c->oper))
       
       { if (c->oper==print_name())
         { if (functype_specialise(type,ctype,ctype))
           return expression_ptr(new  denotation
             (std::make_shared<builtin_value>(print_wrapper,"print")));
         }
         else if (c->oper==prints_name())
         { if (functype_specialise(type,ctype,void_type))
           return expression_ptr(new  denotation
             (std::make_shared<builtin_value>(prints_wrapper,"prints")));
         }
         else if (c->oper==size_of_name())
           
           { if (ctype.kind==row_type)
             { if (functype_specialise(type,ctype,int_type))
             return expression_ptr(new  denotation
                 (std::make_shared<builtin_value>(sizeof_wrapper,"#@[T]")));
               throw type_error(e,ctype.copy(),std::move(type));
             }
             else if (ctype.specialise(pair_type))
             { type_expr& arg_tp0 = *ctype.tupple.begin();
               type_expr& arg_tp1 = *++ctype.tupple.begin();
               if (arg_tp0.kind==row_type)
               { if (arg_tp0==arg_tp1)
                 { if (functype_specialise(type,ctype,arg_tp0))
                   return expression_ptr(new  denotation
                     (std::make_shared<builtin_value>(join_rows_wrapper
                                                     ,"#@([T],[T]->[T])")));
                   throw type_error(e,ctype.copy(),std::move(type));
                 }
                 else if (*arg_tp0.component_type==arg_tp1)
                 { if (functype_specialise(type,ctype,arg_tp0))
                   return expression_ptr(new  denotation
                       (std::make_shared<builtin_value>(suffix_element_wrapper
                                                         ,"#@([T],T->[T])")));
                   throw type_error(e,ctype.copy(),std::move(type));
                 }
               }
               if (arg_tp1.kind==row_type and *arg_tp1.component_type==arg_tp0)
               { if (functype_specialise(type,ctype,arg_tp1))
                 return expression_ptr(new  denotation
                   (std::make_shared<builtin_value>(prefix_element_wrapper
                                                     ,"#@(T,[T]->[T])")));
                 throw type_error(e,ctype.copy(),std::move(type));
               }
             }
           }
       }
     for (size_t i=0; i<variants.size(); ++i)
       if (variants[i].type().arg_type==ctype)
       {
         expression_ptr p(new denotation(variants[i].val));
         const type_expr& res_t = variants[i].type().result_type;
         if (functype_specialise(type,ctype,res_t))
           return p;
    // coercions never apply to values of function type, so just fail here
         throw type_error(e,type_expr(ctype.copy(),res_t.copy()),std::move(type));
       }
     std::ostringstream o;
     o << "Cannot resolve " << main_hash_table->name_of(c->oper) 
       << " at argument type " << ctype;
     throw program_error(o.str());
   }
   break;
   
   case ass_stat:
   {
     id_type lhs=e.assign_variant->lhs;
     const_type_p id_t; size_t i,j;
     const bool is_local = (id_t=layer::lookup(lhs,i,j))!=nullptr;
     if (not is_local and (id_t=global_id_table->type_of(lhs))==nullptr)
       throw program_error  (std::string("Undefined identifier in assignment: ")
             +main_hash_table->name_of(lhs));
   
   type_expr rhs_type = id_t->copy(); // provide a modifiable copy
     expression_ptr r(convert_expr(e.assign_variant->rhs,rhs_type));
     if (rhs_type!=*id_t)
       // assignment will specialise identifier, record to which type it does
     { if (is_local)
         layer::specialise(i,j,rhs_type);
         else
         global_id_table->specialise(lhs,rhs_type);
     }
   
     expression_ptr assign = is_local
     ? expression_ptr(new local_assignment(lhs,i,j,std::move(r)))
   : expression_ptr(new global_assignment(lhs,std::move(r)));
     return conform_types(rhs_type,type,std::move(assign),e);
   }
   
   case comp_ass_stat:
   { id_type aggr=e.comp_assign_variant->aggr;
     const expr& index=e.comp_assign_variant->index;
     const expr& rhs=e.comp_assign_variant->rhs;
   const_type_p aggr_t; size_t d,o;
     bool is_local = (aggr_t=layer::lookup(aggr,d,o))!=nullptr;
     if (not is_local and (aggr_t=global_id_table->type_of(aggr))==nullptr)
       throw program_error 
       (std::string("Undefined identifier in component assignment: ")
        +main_hash_table->name_of(aggr));
   
   
     type_expr ind_t;
     expression_ptr i = convert_expr(index,ind_t);
   type_expr comp_t; subscr_base::sub_type kind;
     if (subscr_base::indexable(*aggr_t,ind_t,comp_t,kind)
         and subscr_base::assignable(kind))
     { expression_ptr r = convert_expr(rhs,comp_t);
       if (aggr_t->kind==row_type)
         aggr_t->component_type->specialise(comp_t); // record type
       if (is_local)
         return conform_types(comp_t,type,expression_ptr(new 
           local_component_assignment(aggr,std::move(i),d,o,std::move(r),kind))
         ,e);
       else
         return conform_types(comp_t,type,expression_ptr(new 
           global_component_assignment(aggr,std::move(i),std::move(r),kind))
         ,e);
     }
     else
     { std::ostringstream o;
       o << "Cannot subscript value of type " << *aggr_t 
         << " with index of type " << ind_t << " in assignment";
       throw expr_error(e,o.str());
     }
   }
   case no_expr: assert(false);
  }
  return nullptr; // keep compiler happy
}

void tuple_expression::print(std::ostream& out) const
{ out << '(';
  for (auto it=component.begin(); it!=component.end(); ++it)
    out << (it==component.begin() ? "" : ",") << **it;
  out << ')';
}

void tuple_expression::evaluate(level l) const
{ if (l==no_value)
    for (auto it=component.begin(); it!=component.end(); ++it)
      (*it)->void_eval();
  else
  { for (auto it=component.begin(); it!=component.end(); ++it)
      (*it)->eval();
    if (l==single_value)
      wrap_tuple(component.size());
  }
}

void list_expression::print(std::ostream& out) const
{ out << '[';
  for (auto it=component.begin(); it!=component.end(); ++it)
    out << (it==component.begin() ? "" : ",") << **it;
  out << ']';
}

void list_expression::evaluate(level l) const
{ if (l==no_value)
    for (auto it=component.begin(); it!=component.end(); ++it)
      (*it)->void_eval();
  else
  { own_row result= std::make_shared<row_value>(0);
    result->val.reserve(component.size());
    for (auto it=component.begin(); it!=component.end(); ++it)
  {
      (*it)->eval();
      result->val.push_back(pop_value());
    }
    push_value(std::move(result)); // result will be shared from here on
  }
}

const char* identifier::name() const
{ return main_hash_table->name_of(code); }
void identifier::print(std::ostream& out) const
{ out<< name(); }

global_identifier::global_identifier(id_type id)
: identifier(id), address(global_id_table->address_of(id))
{}

void global_identifier::evaluate(level l) const
{ if (address->get()==nullptr)
  { std::ostringstream o;
    o << "Taking value of uninitialized variable " << name();
    throw std::runtime_error(o.str());
  }
  push_expanded(l,*address);
}

void local_identifier::evaluate(level l) const
{ push_expanded(l,frame::current->elem(depth,offset)); }

expression_ptr resolve_overload
  (const expr& e,
   type_expr& type,
   const overload_table::variant_list& variants)
{ const expr& args = e.call_variant->arg;
  type_expr a_priori_type;
  expression_ptr arg = convert_expr(args,a_priori_type);
    // get \foreign{a priori} types once
  id_type id =  e.call_variant->fun.identifier_variant;
  
  { if (id==size_of_name())
    { if (a_priori_type.kind==row_type)
      { expression_ptr
          call(new overloaded_builtin_call(sizeof_wrapper,"#",std::move(arg)));
        return conform_types(int_type,type,std::move(call),e);
      }
      else if (a_priori_type.kind!=undetermined_type and
               a_priori_type.specialise(pair_type))
      
      { type_expr& arg_tp0 = *a_priori_type.tupple.begin();
        type_expr& arg_tp1 = *++a_priori_type.tupple.begin();
        if (arg_tp0.kind==row_type)
        { if (can_coerce_arg(arg.get(),1,arg_tp1,*arg_tp0.component_type)) // suffix
          { expression_ptr call(new 
              overloaded_builtin_call (suffix_element_wrapper,"#",std::move(arg)));
            return conform_types(arg_tp0,type,std::move(call),e);
          }
          if (arg_tp0==arg_tp1) // join
          { expression_ptr call(new  overloaded_builtin_call
              (join_rows_wrapper,"#",std::move(arg)));
            return conform_types(arg_tp0,type,std::move(call),e);
          }
        }
        if (arg_tp1.kind==row_type and 
               can_coerce_arg(arg.get(),0,arg_tp0,*arg_tp1.component_type))
                // prefix
        { expression_ptr call(new  overloaded_builtin_call
            (prefix_element_wrapper,"#",std::move(arg)));
          return conform_types(arg_tp1,type,std::move(call),e);
        }
      }
    }
    else if (id==print_name()) // this one always matches
    { expression_ptr call(new
        generic_builtin_call(print_wrapper,"print",std::move(arg)));
      return conform_types(a_priori_type,type,std::move(call),e);
   }
    else if(id==prints_name()) // this always matches as well
    { expression_ptr call(new
        generic_builtin_call(prints_wrapper,"prints",std::move(arg)));
      if (type.specialise(void_type))
        return call;
      throw type_error(e,void_type.copy(),std::move(type));
    }
  }
  for (auto it=variants.begin(); it!=variants.end(); ++it)
  { const overload_data& v=*it;
    if ((is_close(a_priori_type,v.type().arg_type)&0x1)!=0)
      // could first convert to second?
    { if (a_priori_type!=v.type().arg_type)
      { type_expr arg_t = v.type().arg_type.copy(); // need a modifiable value
        arg=convert_expr(args,arg_t); // redo conversion
      }
      
      { expression_ptr call;
        const builtin_value* f = dynamic_cast<const builtin_value*>(v.val.get());
        if (f!=nullptr)
          call = expression_ptr (new 
            overloaded_builtin_call(f->val,f->print_name.c_str(),std::move(arg)));
        else
          
          { shared_closure fun = std::dynamic_pointer_cast<const closure_value>(v.val);
            if (fun==nullptr)
              throw std::logic_error("Overloaded value is not a function");
            std::ostringstream name;
            name << main_hash_table->name_of(id) << '@' << v.type().arg_type;
            call =
              expression_ptr(new overloaded_closure_call(fun,name.str(),std::move(arg)));
          }
        if (type==void_type and
            id==equals_name() and
            v.type().result_type!=void_type)
          throw type_error(e,v.type().result_type.copy(),std::move(type));
        return conform_types(v.type().result_type,type,std::move(call),e);
      }
    }
  }

  
  if (variants.size()==1)
    throw type_error(args,a_priori_type.copy(),variants[0].type().arg_type.copy());
  else
  { std::ostringstream o;
    o << "Failed to match `"
      << main_hash_table->name_of(id) 
      << "' with argument type "
      << a_priori_type;
    throw expr_error(e,o.str());
  }
}

void call_expression::print(std::ostream& out) const
{ if (dynamic_cast<identifier*>(function.get())!=nullptr)
    out << *function;
  else out << '(' << *function << ')';
  if (dynamic_cast<tuple_expression*>(argument.get())!=nullptr)
    out << *argument;
  else out << '(' << *argument << ')';
}

void overloaded_builtin_call::print(std::ostream& out) const
{ out << print_name;
  if (dynamic_cast<tuple_expression*>(argument.get())!=nullptr)
    out << *argument;
  else out << '(' << *argument << ')';
}

void call_expression::evaluate(level l) const
{ function->eval();  shared_value fun=pop_value();
const builtin_value* f=dynamic_cast<const builtin_value*>(fun.get());
  argument->evaluate(f==nullptr ? single_value : multi_value);
  try
  { if (f==nullptr)
      
      { const closure_value* f=force<closure_value>(fun.get());
      
        lambda_frame fr(*f->param,f->context); // save context, create new one for |f|
        fr.bind(pop_value()); // decompose arguments(s) and bind values in |fr|
        f->body->evaluate(l); // call, passing evaluation level |l| to function body
      } 
    else // built-in functions
      (*f->val)(l); // call the wrapper function, handling |l| appropriately
  }
  
  catch (const std::exception& e)
  { identifier* p=dynamic_cast<identifier*>(function.get());
    if (p!=nullptr) // named function
    {
      const std::logic_error* l_err= dynamic_cast<const std::logic_error*>(&e);
      if (l_err!=nullptr)
        throw std::logic_error
          (std::string(e.what())+"\n(in call of "+p->name()+')');
      throw std::runtime_error
        (std::string(e.what())+"\n(in call of "+p->name()+')');
    }
    throw; // for anonymous function calls, just rethrow the error unchanged
  }
}

void overloaded_builtin_call::evaluate(level l) const
{ argument->multi_eval();
  try
  { (*f)(l); }
  catch (const std::exception& e)
  { const std::logic_error* l_err= dynamic_cast<const std::logic_error*>(&e);
    if (l_err!=nullptr)
      throw std::logic_error
        (std::string(e.what())+"\n(in call of "+print_name+')');
    throw std::runtime_error
      (std::string(e.what())+"\n(in call of "+print_name+')');
  }
}

void generic_builtin_call::evaluate(level l) const
{ argument->eval();
  try
  { (*f)(l); }
  catch (const std::exception& e)
  { const std::logic_error* l_err= dynamic_cast<const std::logic_error*>(&e);
    if (l_err!=nullptr)
      throw std::logic_error
        (std::string(e.what())+"\n(in call of "+print_name+')');
    throw std::runtime_error
      (std::string(e.what())+"\n(in call of "+print_name+')');
  }
}

inline
let_expression::let_expression 
  (const id_pat& v, expression_ptr&& ini, expression_ptr&& b)
: variable(copy_id_pat(v))
, initialiser(ini.release())
, body(b.release())
{}

void let_expression::print(std::ostream& out) const
{ out << " let " << variable << "=" << *initialiser << " in " << *body;
}

type_list pattern_list(const patlist& p)
{ dressed_type_list result;
  for (auto it = p.begin(); not p.at_end(it); ++it)
    result.push_back(pattern_type(*it));
  return result.undress();
}

type_expr pattern_type(const id_pat& pat)
{ return (pat.kind&0x2)==0
  ? unknown_type.copy()
  : type_expr(pattern_list(pat.sublist));
}

size_t count_identifiers(const id_pat& pat)
{ size_t result= pat.kind & 0x1; // 1 if |pat.name| is defined, 0 otherwise
  if ((pat.kind & 0x2)!=0) // then a list of subpatterns is present
    for (auto it=pat.sublist.begin(); not pat.sublist.at_end(it); ++it)
      result+=count_identifiers(*it);
  return result;
}

void list_identifiers(const id_pat& pat, std::vector<id_type>& d)
{ if ((pat.kind & 0x1)!=0)
    d.push_back(pat.name);
  if ((pat.kind & 0x2)!=0) // then a list of subpatterns is present
    for (auto it=pat.sublist.begin(); not pat.sublist.at_end(it); ++it)
      list_identifiers(*it,d);
}

void thread_bindings
(const id_pat& pat,const type_expr& type, layer& dst)
{ if ((pat.kind & 0x1)!=0) dst.add(pat.name,type.copy());
  if ((pat.kind & 0x2)!=0)
  { assert(type.kind==tuple_type);
    type_list::const_iterator t_it = type.tupple.begin();
    for (auto p_it=pat.sublist.begin(); not pat.sublist.at_end(p_it);
         ++p_it,++t_it)
      thread_bindings(*p_it,*t_it,dst);
  }
}

void thread_components
  (const id_pat& pat,const shared_value& val
  , std::back_insert_iterator<std::vector<shared_value> > dst)
{ if ((pat.kind & 0x1)!=0)
     *dst++ = val; // copy |shared_value| pointer |val|, creating sharing

  if ((pat.kind & 0x2)!=0)
  { const tuple_value* t=force<tuple_value>(val.get());
    assert(t->val.size()==length(pat.sublist));
    size_t i=0;
    for (auto it=pat.sublist.begin(); not pat.sublist.at_end(it); ++it,++i)
      thread_components(*it,t->val[i],dst);
  }
}

void let_expression::evaluate(level l) const
{ initialiser->eval(); // evaluate on stack as single value

  frame fr(variable); // save context, create new one for |f|
  fr.bind(pop_value()); // decompose arguments(s) and bind values in |fr|
  body->evaluate(l); // call, passing evaluation level |l| to function body
} 

inline
lambda_expression::lambda_expression(const id_pat& p, expression_ptr&& b)
: param(std::make_shared<id_pat>(copy_id_pat(p)))
, body(b.release())
{}

void lambda_expression::print(std::ostream& out) const
{ if ((param->kind&0x1)!=0)
    out << '(' << *param << ')';
  else out << *param;
  out << ": " << *body;
}

void closure_value::print(std::ostream& out) const
{ if ((param->kind&0x1)!=0)
    out << '(' << *param << ')';
  else out << *param;
  out << ": " << *body;
}

void lambda_expression::evaluate(level l) const
{if (l!=no_value)
     push_value(std::make_shared<closure_value>(frame::current,param,body));
}

void overloaded_closure_call::print(std::ostream& out) const
{ out << print_name;
  if (dynamic_cast<tuple_expression*>(argument.get())!=nullptr)
     out << *argument;
  else out << '(' << *argument << ')';
}

void overloaded_closure_call::evaluate(level l) const
{ argument->eval();
  try
  { lambda_frame fr(*fun->param,fun->context);
    // save context, create new one for |fun|
    fr.bind(pop_value()); // decompose arguments(s) and bind values in |fr|
    fun->body->evaluate(l);
    // call, passing evaluation level |l| to function body
  }
  catch (const std::exception& e)
  { const std::logic_error* l_err= dynamic_cast<const std::logic_error*>(&e);
    if (l_err!=nullptr)
      throw std::logic_error
        (std::string(e.what())+"\n(in call of "+print_name+')');
    throw std::runtime_error
      (std::string(e.what())+"\n(in call of "+print_name+')');
  }
}

void seq_expression::print(std::ostream& out) const
{ out << *first << ';' << *last; }
void next_expression::print(std::ostream& out) const
{ out << *first << " next " << *last; }

void seq_expression::evaluate(level l) const
{ first->void_eval(); last->evaluate(l); }

void next_expression::evaluate(level l) const
{ first->evaluate(l); last->void_eval(); }

void subscr_base::print(std::ostream& out) const
{ out << *array << '[' << *index << ']';
}

void matrix_subscription::print(std::ostream& out) const
{ tuple_expression* p=dynamic_cast<tuple_expression*>(index.get());
  if (p==nullptr) out << *array << '[' << *index << ']';
  else
    out << *array << '[' << *p->component[0] << ',' << *p->component[1] << ']';
}

bool subscr_base::indexable
  (const type_expr& aggr,
   const type_expr& index,
         type_expr& subscr,
         sub_type& kind)
{ if (aggr.kind==primitive_type)
  switch (aggr.prim)
  {  default:
  break; case vector_type: if (index==int_type)
    { kind=vector_entry; return subscr.specialise(int_type);  }
  break; case rational_vector_type: if(index==int_type)
    { kind=ratvec_entry; return subscr.specialise(rat_type); }
  break; case string_type: if(index==int_type)
    { kind=string_char; return subscr.specialise(str_type); }
  break; case matrix_type:
    if (index==int_int_type)
    { kind=matrix_entry; return subscr.specialise(int_type); }
    else if (index==int_type)
    { kind=matrix_column; return subscr.specialise(vec_type); }
  break; case virtual_module_type: if (index==param_type)
    { kind=mod_poly_term; return subscr.specialise(split_type); }
  }
  else if (aggr.kind==row_type and index==int_type)
  { kind=row_entry;
        return subscr.specialise(*aggr.component_type);
  }
  return false;
}

bool subscr_base::assignable(subscr_base::sub_type t)
{ switch (t)
  { case ratvec_entry: case string_char: case mod_poly_term: return false;
    default: return true;
  }
}

inline std::string range_mess(int i,size_t n,const expression_base* e)
{ std::ostringstream o;
  e->print(o << "index " << i << " out of range (<" << n
             << ") in subscription ");
  return o.str();
}

void row_subscription::evaluate(level l) const
{ shared_int i=(index->eval(),get<int_value>());
  shared_row r=(array->eval(),get<row_value>());
  if (static_cast<unsigned int>(i->val)>=r->val.size())
    throw std::runtime_error(range_mess(i->val,r->val.size(),this));
  push_expanded(l,r->val[i->val]);
}

void vector_subscription::evaluate(level l) const
{ shared_int i=(index->eval(),get<int_value>());
  shared_vector v=(array->eval(),get<vector_value>());
  if (static_cast<unsigned int>(i->val)>=v->val.size())
    throw std::runtime_error(range_mess(i->val,v->val.size(),this));
  if (l!=no_value)
    push_value(std::make_shared<int_value>(v->val[i->val]));
}

void ratvec_subscription::evaluate(level l) const
{ shared_int i=(index->eval(),get<int_value>());
  shared_rational_vector v=(array->eval(),get<rational_vector_value>());
  if (static_cast<unsigned int>(i->val)>=v->val.size())
    throw std::runtime_error(range_mess(i->val,v->val.size(),this));
  if (l!=no_value)
    push_value(std::make_shared<rat_value>(Rational 
       (v->val.numerator()[i->val],v->val.denominator())));
}

void string_subscription::evaluate(level l) const
{ shared_int i=(index->eval(),get<int_value>());
  shared_string s=(array->eval(),get<string_value>());
  if (static_cast<unsigned int>(i->val)>=s->val.size())
    throw std::runtime_error(range_mess(i->val,s->val.size(),this));
  if (l!=no_value)
    push_value(std::make_shared<string_value>(s->val.substr(i->val,1)));
}

void matrix_subscription::evaluate(level l) const
{ index->multi_eval(); 
  shared_int j=get<int_value>();
  shared_int i=get<int_value>();
  shared_matrix m=(array->eval(),get<matrix_value>());
  if (static_cast<unsigned int>(i->val)>=m->val.numRows())
    throw std::runtime_error
     ("initial "+range_mess(i->val,m->val.numRows(),this));
  if (static_cast<unsigned int>(j->val)>=m->val.numColumns())
    throw std::runtime_error
     ("final "+range_mess(j->val,m->val.numColumns(),this));
  if (l!=no_value)
    push_value(std::make_shared<int_value>(m->val(i->val,j->val)));
}

void matrix_slice::evaluate(level l) const
{ shared_int j=(index->eval(),get<int_value>());
  shared_matrix m=(array->eval(),get<matrix_value>());
  if (static_cast<unsigned int>(j->val)>=m->val.numColumns())
    throw std::runtime_error(range_mess(j->val,m->val.numColumns(),this));
  if (l!=no_value)
    push_value(std::make_shared<vector_value>(m->val.column(j->val)));
}

void conditional_expression::print(std::ostream& out) const
{ out << " if "; const conditional_expression* cur=this;
  do
  { out << *cur->condition << " then " << *cur->then_branch;
    conditional_expression* p =
      dynamic_cast<conditional_expression*>(cur->else_branch.get());
    if (p==nullptr)
      break;
    out << " elif "; cur=p;
  }
  while(true);
  out << " else " << *cur->else_branch << " fi ";
}

void conditional_expression::evaluate(level l) const
{ condition->eval();
  if (get<bool_value>()->val)
    then_branch->evaluate(l);
   else else_branch->evaluate(l);
}

void while_expression::print(std::ostream& out) const
{ out << " while " <<  *condition << " do " << *body << " od ";
}

void while_expression::evaluate(level l) const
{ if (l==no_value)
    while (condition->eval(),get<bool_value>()->val)
       body->void_eval();
  else
  { own_row result = std::make_shared<row_value>(0);
    while (condition->eval(),get<bool_value>()->val)
  { body->eval();
      result->val.push_back(pop_value());
    }
    push_value(std::move(result));
  }
}

inline
for_expression::for_expression
 (const id_pat& p, expression_ptr&& i, expression_ptr&& b
 ,subscr_base::sub_type k)
   : pattern(copy_id_pat(p)), in_part(i.release()), body(b.release()), kind(k)
  {}

void for_expression::print(std::ostream& out) const
{ out << " for " << *++pattern.sublist.begin();
    if (pattern.sublist.front().kind==0x1)
      out << '@' << pattern.sublist.front();
    out << " in " << *in_part << " do " << *body << " od ";
}

void for_expression::evaluate(level l) const
{ in_part->eval();
  own_tuple loop_var = std::make_shared<tuple_value>(2);
       // this is safe to re-use between iterations
  own_row result(nullptr);
  
  switch (kind)
  { case subscr_base::row_entry:
    { shared_row in_val = get<row_value>();
      size_t n=in_val->val.size();
      if (l!=no_value)
        result = std::make_shared<row_value>(n);
      for (size_t i=0; unsigned(i)<n; ++i)
      { loop_var->val[1]=in_val->val[i]; // share the current row component
        
        { loop_var->val[0] = std::make_shared<int_value>(i);
            // index; newly created each time
          frame fr (pattern);
          fr.bind(loop_var);
          if (l==no_value)
            body->void_eval();
          else
          { body->eval();
            result->val[i] = pop_value();
          }
        } 
      }
    }
    break;
    case subscr_base::vector_entry:
    { shared_vector in_val = get<vector_value>();
      size_t n=in_val->val.size();
      if (l!=no_value)
        result = std::make_shared<row_value>(n);
      for (size_t i=0; unsigned(i)<n; ++i)
      { loop_var->val[1] = std::make_shared<int_value>(in_val->val[i]);
        
        { loop_var->val[0] = std::make_shared<int_value>(i);
            // index; newly created each time
          frame fr (pattern);
          fr.bind(loop_var);
          if (l==no_value)
            body->void_eval();
          else
          { body->eval();
            result->val[i] = pop_value();
          }
        } 
      }
    }
    break;
    case subscr_base::ratvec_entry:
    { shared_rational_vector in_val = get<rational_vector_value>();
      size_t n=in_val->val.size();
      if (l!=no_value)
        result = std::make_shared<row_value>(n);
      for (size_t i=0; unsigned(i)<n; ++i)
      { loop_var->val[1] = std::make_shared<rat_value>(Rational 
          (in_val->val.numerator()[i],in_val->val.denominator()));
        
        { loop_var->val[0] = std::make_shared<int_value>(i);
            // index; newly created each time
          frame fr (pattern);
          fr.bind(loop_var);
          if (l==no_value)
            body->void_eval();
          else
          { body->eval();
            result->val[i] = pop_value();
          }
        } 
      }
    }
    break;
    case subscr_base::string_char:
    { shared_string in_val = get<string_value>();
      size_t n=in_val->val.size();
      if (l!=no_value)
        result = std::make_shared<row_value>(n);
      for (size_t i=0; unsigned(i)<n; ++i)
      { loop_var->val[1] = std::make_shared<string_value>(in_val->val.substr(i,1));
        
        { loop_var->val[0] = std::make_shared<int_value>(i);
            // index; newly created each time
          frame fr (pattern);
          fr.bind(loop_var);
          if (l==no_value)
            body->void_eval();
          else
          { body->eval();
            result->val[i] = pop_value();
          }
        } 
      }
    }
    break;
    case subscr_base::matrix_column:
    { shared_matrix in_val = get<matrix_value>();
      size_t n=in_val->val.numColumns();
      if (l!=no_value)
        result = std::make_shared<row_value>(n);
      for (size_t i=0; unsigned(i)<n; ++i)
      { loop_var->val[1] = std::make_shared<vector_value>(in_val->val.column(i));
        
        { loop_var->val[0] = std::make_shared<int_value>(i);
            // index; newly created each time
          frame fr (pattern);
          fr.bind(loop_var);
          if (l==no_value)
            body->void_eval();
          else
          { body->eval();
            result->val[i] = pop_value();
          }
        } 
      }
    }
    break;
    case subscr_base::mod_poly_term:
    
    { shared_virtual_module pol_val = get<virtual_module_value>();
      size_t n=pol_val->val.size(),i=0;
      if (l!=no_value)
        result = std::make_shared<row_value>(n);
      for (auto it=pol_val->val.begin(); it!=pol_val->val.end(); ++it,++i)
      { loop_var->val[0] =
          std::make_shared<module_parameter_value>(pol_val->rf,it->first);
        loop_var->val[1] = std::make_shared<split_int_value>(it->second);
        frame fr(pattern);
        fr.bind(loop_var);
        if (l==no_value)
          body->void_eval();
        else
        { body->eval();
          result->val[i]=pop_value();
        }
      } // restore context upon destruction of |fr|
    }
    break;
    case subscr_base::matrix_entry: break; // excluded in type analysis
  }
  if (l!=no_value)
    push_value(std::move(result));
}

void inc_for_expression::print(std::ostream& out) const
{ out << " for " << main_hash_table->name_of(id)  << ": " << *count 
      << " from " << *bound << " do " << *body << " od ";
}
void dec_for_expression::print(std::ostream& out) const
{ out << " for " << main_hash_table->name_of(id)  << " : " << *count 
      << " downto " << *bound << " do " << *body << " od ";
}

void inc_for_expression::evaluate(level l) const
{ int b=(bound->eval(),get<int_value>()->val);
  int c=(count->eval(),get<int_value>()->val);
  if (c<0)
    c=0; // no negative size result

  id_pat pattern(id,0x1,patlist(nullptr));
  if (l==no_value)
  { c+=b;
    for (int i=b; i<c; ++i)
  { frame fr(pattern);
      fr.bind(std::make_shared<int_value>(i));
      body->void_eval();
    }
  }
  else
  { own_row result = std::make_shared<row_value>(0); result->val.reserve(c);
    c+=b;
    for (int i=b; i<c; ++i)
    { frame fr(pattern);
      fr.bind(std::make_shared<int_value>(i));
      body->eval();
      result->val.push_back(pop_value());
    }
    push_value(std::move(result));
  }
}

void dec_for_expression::evaluate(level l) const
{ int b=(bound->eval(),get<int_value>()->val);
  int i=(count->eval(),get<int_value>()->val);
  if (i<0)
    i=0; // no negative size result

  id_pat pattern(id,0x1,patlist(nullptr));
  if (l==no_value)
  { i+=b;
    while (i-->b)
  { frame fr(pattern);
      fr.bind(std::make_shared<int_value>(i));
      body->void_eval();
    }
  }
  else
  { own_row result = std::make_shared<row_value>(0); result->val.reserve(i);
    i+=b;
    while (i-->b)
    { frame fr(pattern);
      fr.bind(std::make_shared<int_value>(i));
      body->eval();
      result->val.push_back(pop_value());
    }
    push_value(std::move(result));
  }
}

void assignment_expr::print(std::ostream& out) const
{ out << main_hash_table->name_of(lhs) << ":=" << *rhs; }

global_assignment::global_assignment(id_type l,expression_ptr&& r)
: assignment_expr(l,std::move(r)), address(global_id_table->address_of(l)) {}

void global_assignment::evaluate(level l) const
{ rhs->eval();
  *address = pop_value();
  push_expanded(l,*address);
}

local_assignment::local_assignment
 (id_type l, size_t i,size_t j, expression_ptr&& r)
: assignment_expr(l,std::move(r)), depth(i), offset(j) {}

void local_assignment::evaluate(level l) const
{ rhs->eval();
  shared_value& dest =  frame::current->elem(depth,offset);
  dest= pop_value();
  push_expanded(l,dest);
}

void component_assignment::print(std::ostream& out) const
{ out << main_hash_table->name_of(lhs) << '[' << *index << "]:=" << *rhs; }

global_component_assignment::global_component_assignment 
  (id_type a,expression_ptr&& i,expression_ptr&& r,
   subscr_base::sub_type k)
: component_assignment(a,std::move(i),std::move(r))
, kind(k),address(global_id_table->address_of(a)) {}

void global_component_assignment::evaluate(level l) const
{ if (address->get()==nullptr)
  { std::ostringstream o;
    o << "Assigning to component of uninitialized variable "
      << main_hash_table->name_of(lhs);
    throw std::runtime_error(o.str());
  }
  assign(l,*address,kind);
}

void component_assignment::assign
  (level l,shared_value& aggregate, subscr_base::sub_type kind) const
{ rhs->eval();
  value loc=uniquify(aggregate);
    // simple pointer to modifiable value from shared pointer
  switch (kind)
  { case subscr_base::row_entry:
  
  { unsigned int i=(index->eval(),get<int_value>()->val);
    std::vector<shared_value>& a=force<row_value>(loc)->val;
    if (i>=a.size())
      throw std::runtime_error(range_mess(i,a.size(),this));
    a[i] = pop_value();
    push_expanded(l,a[i]);
  }
  break;
    case subscr_base::vector_entry:
  
  { unsigned int i=(index->eval(),get<int_value>()->val);
    std::vector<int>& v=force<vector_value>(loc)->val;
    if (i>=v.size())
      throw std::runtime_error(range_mess(i,v.size(),this));
    v[i]= force<int_value>(execution_stack.back().get())->val;
    if (l==no_value)
      execution_stack.pop_back();
  }
  break;
    case subscr_base::matrix_entry:
  
  { index->multi_eval();
    unsigned int j=get<int_value>()->val;
    unsigned int i=get<int_value>()->val;
  
    int_Matrix& m=force<matrix_value>(loc)->val;
    if (i>=m.numRows())
      throw std::runtime_error(range_mess(i,m.numRows(),this));
    if (j>=m.numColumns())
      throw std::runtime_error(range_mess(j,m.numColumns(),this));
    m(i,j)= force<int_value>(execution_stack.back().get())->val;
    if (l==no_value)
      execution_stack.pop_back();
  }
  break;
    case subscr_base::matrix_column:
  
  { unsigned int j=(index->eval(),get<int_value>()->val);
    int_Matrix& m=force<matrix_value>(loc)->val;
    const int_Vector& v=force<vector_value>(execution_stack.back().get())->val;
    if (j>=m.numColumns())
      throw std::runtime_error(range_mess(j,m.numColumns(),this));
    if (v.size()!=m.numRows())
      throw std::runtime_error
        (std::string("Cannot replace column of size ")+str(m.numRows())+
         " by one of size "+str(v.size()));
    m.set_column(j,v);
    if (l==no_value)
      execution_stack.pop_back();
  }
  break;
  default: {} // remaining cases are eliminated in type analysis
  }
}

local_component_assignment::local_component_assignment
 (id_type l, expression_ptr&& i,size_t d, size_t o, expression_ptr&& r,
  subscr_base::sub_type k)
: component_assignment(l,std::move(i),std::move(r))
, kind(k), depth(d), offset(o) {}

void local_component_assignment::evaluate(level l) const
{ assign(l,frame::current->elem(depth,offset),kind); }

} }

