
#ifndef EVALUATOR_H
#define EVALUATOR_H

#include "types.h"


#include "global.h" 

namespace atlas { namespace interpreter {

class layer
{
  typedef containers::simple_list<layer*> list;
public:
  typedef std::vector<std::pair<id_type,type_expr> > vec;
  static list lexical_context; // the unique |layer::list| in existence
private:
  vec variable;
public:
  layer(const layer&) = delete; // no ordinary copy constructor
  layer& operator= (const layer&) = delete; // nor assignment operator
  layer(size_t n) : variable()
  { variable.reserve(n); lexical_context.push_front(this); }
  ~layer () { lexical_context.pop_front(); }

  void add(id_type id,type_expr&& t);
  static const_type_p lookup
    (id_type id, size_t& depth, size_t& offset);
  static void specialise (size_t depth, size_t offset,const type_expr& t);

  std::pair<id_type,type_expr>& operator[] (size_t i)
  { return variable[i]; }
  vec::iterator begin() { return variable.begin(); }
  vec::iterator end() { return variable.end(); }
  vec::const_iterator cbegin() { return variable.begin(); }
  vec::const_iterator cend() { return variable.end(); }
};

struct denotation : public expression_base
{ shared_value denoted_value;

  explicit denotation(const shared_value& v) : denoted_value(v) {}
  explicit denotation(shared_value&& v) : denoted_value(std::move(v)) {}
  virtual void evaluate(level l) const { push_expanded(l,denoted_value); }
  virtual void print(std::ostream& out) const
  { denoted_value->print(out); }
};

struct tuple_expression : public expression_base
{ std::vector<expression_ptr> component;

  explicit tuple_expression(size_t n) : component(n) {}
   // always start out with null pointers
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct list_expression : public tuple_expression
{ explicit list_expression(size_t n) : tuple_expression(n){}
virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

#define nothing_new_here {}
  // patch for gcc 4.6, will become |=default| in \Cpp11

struct identifier : public expression_base
{ id_type code;

  explicit identifier(id_type id) : code(id) {}
  virtual ~identifier() nothing_new_here;
  const char* name() const;
  virtual void print(std::ostream& out) const;
};

class global_identifier : public identifier
{ const shared_share address;
public:
  explicit global_identifier(id_type id);
  virtual ~global_identifier() nothing_new_here;
  virtual void evaluate(level l) const;
};

class local_identifier : public identifier
{ size_t depth, offset;
public:
  explicit local_identifier(id_type id, size_t i, size_t j)
     : identifier(id), depth(i), offset(j) {}
  virtual void evaluate(level l) const; // only this method is redefined
};

struct call_expression : public expression_base
{ expression_ptr function, argument;

  call_expression(expression_ptr&& f,expression_ptr&& a)
   : function(f.release()),argument(a.release()) {}
  virtual ~call_expression() nothing_new_here;
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct builtin_value : public value_base
{ wrapper_function val;
  std::string print_name;

  builtin_value(wrapper_function v,const std::string& n)
  : val(v), print_name(n)  {}
  virtual ~ builtin_value() nothing_new_here;
  virtual void print(std::ostream& out) const
  { out << '{' << print_name << '}'; }
  virtual builtin_value* clone() const { return new builtin_value(*this); }
  static const char* name() { return "built-in function"; }
private:
  builtin_value(const builtin_value& v) = default;
};

struct overloaded_builtin_call : public expression_base
{ wrapper_function f;
  std::string print_name;
  expression_ptr argument;

  overloaded_builtin_call(wrapper_function v,const char* n,expression_ptr&& a)
  : f(v), print_name(n), argument(a.release()) {}
  virtual ~overloaded_builtin_call() nothing_new_here;
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct generic_builtin_call : public overloaded_builtin_call
{ typedef overloaded_builtin_call base;

  generic_builtin_call(wrapper_function v,const char* n,expression_ptr&& a)
  : base(v,n,std::move(a)) {}
  virtual void evaluate(level l) const;
};

struct let_expression : public expression_base
{ const id_pat variable;
  expression_ptr initialiser, body;

  let_expression(const id_pat& v, expression_ptr&& ini, expression_ptr&& b);
  virtual ~let_expression() nothing_new_here; // subobjects do all the work
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

typedef std::shared_ptr<id_pat> shared_pattern;

struct lambda_expression : public expression_base
{ const shared_pattern param;
  shared_expression body;

  lambda_expression(const id_pat& p, expression_ptr&& b);
  virtual ~lambda_expression() nothing_new_here; // subobjects do all the work
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct closure_value : public value_base
{ shared_context context;
  shared_pattern param;
   // used in evaluation only to count arguments
  shared_expression body;

  closure_value(const shared_context& c,
                  const shared_pattern& p,
                  const shared_expression& b)
  : context(c), param(p), body(b) {}
  virtual ~ closure_value() nothing_new_here;
  virtual void print(std::ostream& out) const;
  virtual closure_value* clone() const 
  { return new closure_value(context,param,body); }
  static const char* name() { return "closure"; }
};
typedef std::unique_ptr<closure_value> closure_ptr;
typedef std::shared_ptr<const closure_value> shared_closure;

struct overloaded_closure_call : public expression_base
{ shared_closure fun;
  std::string print_name;
  expression_ptr argument;

  overloaded_closure_call
   (shared_closure f,const std::string& n,expression_ptr&& a)
  : fun(f), print_name(n), argument(a.release()) {}
  virtual ~overloaded_closure_call() nothing_new_here;
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct seq_expression : public expression_base
{ expression_ptr first,last;

  seq_expression(expression_ptr&& f,expression_ptr&& l)
   : first(f.release()),last(l.release()) {}
  virtual ~seq_expression() nothing_new_here;
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct next_expression : public expression_base
{ expression_ptr first,last;

  next_expression(expression_ptr&& f,expression_ptr&& l)
   : first(f.release()),last(l.release()) {}
  virtual ~next_expression() nothing_new_here;
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct subscr_base : public expression_base
{ enum sub_type
  { row_entry, vector_entry, ratvec_entry, string_char
  , matrix_entry, matrix_column, mod_poly_term };
  expression_ptr array, index; // the two parts of the subscription expression

  subscr_base(expression_ptr&& a, expression_ptr&& i)
: array(a.release()),index(i.release()) {}
  virtual ~subscr_base() nothing_new_here ;

  virtual void print(std::ostream& out) const;
  static bool indexable
  (const type_expr& aggr,
   const type_expr& index,
         type_expr& subscr,
         sub_type& kind);
  static bool assignable(sub_type);
};

struct row_subscription : public subscr_base
{ row_subscription(expression_ptr&& a, expression_ptr&& i)
: subscr_base(std::move(a),std::move(i)) {}
  virtual void evaluate(level l) const;
};


struct vector_subscription : public subscr_base
{ vector_subscription(expression_ptr&& a, expression_ptr&& i)
: subscr_base(std::move(a),std::move(i)) {}
  virtual void evaluate(level l) const;
};

struct ratvec_subscription : public subscr_base
{ ratvec_subscription(expression_ptr&& a, expression_ptr&& i)
: subscr_base(std::move(a),std::move(i)) {}
  virtual void evaluate(level l) const;
};

struct string_subscription : public subscr_base
{ string_subscription(expression_ptr&& a, expression_ptr&& i)
: subscr_base(std::move(a),std::move(i)) {}
  virtual void evaluate(level l) const;
};

struct matrix_subscription : public subscr_base
{ matrix_subscription(expression_ptr&& a, expression_ptr&& ij)
: subscr_base(std::move(a),std::move(ij)) {}
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct matrix_slice : public subscr_base
{ matrix_slice(expression_ptr&& a, expression_ptr&& j)
: subscr_base(std::move(a),std::move(j)) {}
  virtual void evaluate(level l) const;
};

struct module_coefficient : public subscr_base
{ module_coefficient(expression_ptr&& pol, expression_ptr&& param)
: subscr_base(std::move(pol),std::move(param)) {}
  virtual void evaluate(level l) const;
};

struct conditional_expression : public expression_base
{ expression_ptr condition, then_branch, else_branch;

  conditional_expression
   (expression_ptr&& c,expression_ptr&& t, expression_ptr&& e)
   : condition(c.release()),then_branch(t.release()), else_branch(e.release())
  {}
  virtual ~conditional_expression() nothing_new_here;
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct while_expression : public expression_base
{ expression_ptr condition, body;

  while_expression(expression_ptr&& c,expression_ptr&& b)
   : condition(c.release()),body(b.release())
  {}
  virtual ~while_expression() nothing_new_here;
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct for_expression : public expression_base
{ id_pat pattern; expression_ptr in_part, body; subscr_base::sub_type kind;

  for_expression
   (const id_pat& p, expression_ptr&& i, expression_ptr&& b
   , subscr_base::sub_type k);
  virtual ~for_expression() nothing_new_here;
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct inc_for_expression : public expression_base
{ expression_ptr count, bound, body; id_type id;

  inc_for_expression
   (id_type i, expression_ptr&& cnt, expression_ptr&& bnd,
    expression_ptr&& b)
  : count(cnt.release()),bound(bnd.release()),body(b.release()),id(i)
  {}
  virtual ~inc_for_expression() nothing_new_here;
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct dec_for_expression : public expression_base
{ expression_ptr count, bound, body; id_type id;

  dec_for_expression
   (id_type i, expression_ptr&& cnt, expression_ptr&& bnd,
    expression_ptr&& b)
  : count(cnt.release()),bound(bnd.release()),body(b.release()),id(i)
  {}
  virtual ~dec_for_expression() nothing_new_here;
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct assignment_expr : public expression_base
{ id_type lhs;
  expression_ptr rhs;

  assignment_expr(id_type l,expression_ptr&& r)
   : lhs(l),rhs(r.release()) {}
  virtual ~assignment_expr() nothing_new_here;
  virtual void print(std::ostream& out) const;
};

class global_assignment : public assignment_expr
{ shared_share address;
public:
  global_assignment(id_type l,expression_ptr&& r);
  virtual ~global_assignment() nothing_new_here;
  virtual void evaluate(level l) const;
};

class local_assignment : public assignment_expr
{ size_t depth, offset;
public:
  local_assignment(id_type l, size_t i,size_t j, expression_ptr&& r);
  virtual ~local_assignment() nothing_new_here;
  virtual void evaluate(level l) const;
};

struct component_assignment : public assignment_expr
{ expression_ptr index;

  component_assignment
   (id_type a,expression_ptr&& i,expression_ptr&& r)
   : assignment_expr(a,std::move(r)), index(i.release()) {}
  virtual ~component_assignment() nothing_new_here;

  virtual void print (std::ostream& out) const;

  void assign(level l,shared_value& aggregate,subscr_base::sub_type kind) const;
};

class global_component_assignment : public component_assignment
{ subscr_base::sub_type kind;
  shared_share address;
public:
  global_component_assignment
    (id_type a,expression_ptr&& i,expression_ptr&& r,
     subscr_base::sub_type k);
  virtual void evaluate(level l) const;
};

class local_component_assignment : public component_assignment
{ subscr_base::sub_type kind;
  size_t depth, offset;
public:
  local_component_assignment 
   (id_type l, expression_ptr&& i,size_t d, size_t o,
    expression_ptr&& r, subscr_base::sub_type k);
  virtual void evaluate(level l) const;
};


extern type_expr last_type;
extern shared_value last_value;


void reset_evaluator ();

expression_ptr convert_expr(const expr& e, type_expr& type);

expression_ptr resolve_overload
  (const expr& e,
   type_expr& type,
   const overload_table::variant_list& variants);

type_expr pattern_type(const id_pat& pat);
size_t count_identifiers(const id_pat& pat);
void list_identifiers(const id_pat& pat, std::vector<id_type>& d);
void thread_bindings(const id_pat& pat,const type_expr& type, layer& dst);
void thread_components
  (const id_pat& pat,const shared_value& val,
   std::back_insert_iterator<std::vector<shared_value> > dst);

} }
#endif

