#include  "global.h"
#include  <cstdlib>
#include  "evaluator.h"
#include <sstream>
#include <iomanip>
#include  "bitvector.h"
#include  "matreduc.h"
#include  "bitmap.h"
#include  "lattice.h"
#include  "bitvector.h"
#include  "permutations.h"

namespace atlas { namespace interpreter {

Id_table* global_id_table=nullptr; 

overload_table* global_overload_table=nullptr;

std::ostream* output_stream= &std::cout;

int verbosity=0;

namespace {

void rational_convert() // convert integer to rational (with denominator~1)
{ int i = get<int_value>()->val;
    push_value(std::make_shared<rat_value>(Rational(i)));
}

void ratlist_ratvec_convert() // convert list of rationals to rational vector
{ shared_row r = get<row_value>();
  int_Vector numer(r->val.size()),denom(r->val.size());
  unsigned int d=1;
  for (size_t i=0; i<r->val.size(); ++i)
  { Rational frac = force<rat_value>(r->val[i].get())->val;
    numer[i]=frac.numerator();
    denom[i]=frac.denominator();
    d=arithmetic::lcm(d,denom[i]);
  }
  for (size_t i=0; i<r->val.size(); ++i)
    numer[i]*= d/denom[i]; // adjust numerators to common denominator

  push_value(std::make_shared<rational_vector_value>(numer,d)); // normalises
}

void ratvec_ratlist_convert() // convert rational vector to list of rationals
{ shared_rational_vector rv = get<rational_vector_value>();
  own_row result = std::make_shared<row_value>(rv->val.size());
  for (size_t i=0; i<rv->val.size(); ++i)
  { Rational q(rv->val.numerator()[i],rv->val.denominator());
    result->val[i] = std::make_shared<rat_value>(q.normalize());
  }
  push_value(std::move(result));
}

void vec_ratvec_convert() // convert vector to rational vector
{ shared_vector v = get<vector_value>();
  push_value(std::make_shared<rational_vector_value>(RatWeight(v->val,1)));
}

int_Vector row_to_weight(const row_value& r)
{ int_Vector result(r.val.size());
  for(size_t i=0; i<r.val.size(); ++i)
    result[i]=force<int_value>(r.val[i].get())->val;
  return result;
}

void intlist_vector_convert()
{ shared_row r = get<row_value>();
  push_value(std::make_shared<vector_value>(row_to_weight(*r)));
}

void intlist_ratvec_convert()
{ shared_row r = get<row_value>();
  push_value(std::make_shared<rational_vector_value>
    (RatWeight(row_to_weight(*r),1)));
}

void veclist_matrix_convert()
{ shared_row r = get<row_value>();
  if (r->val.size()==0)
    throw std::runtime_error("Cannot convert empty list of vectors to matrix");

  size_t n = force<vector_value>(r->val[0].get())->val.size();
  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,r->val.size()));
  for(size_t j=0; j<r->val.size(); ++j)
  { const int_Vector& col = force<vector_value>(r->val[j].get())->val;
    if (col.size()!=n)
      throw std::runtime_error("Vector sizes differ in conversion to matrix");

    m->val.set_column(j,col);
  }
  push_value(std::move(m));
}

void intlistlist_matrix_convert()
{ shared_row r = get<row_value>();
  if (r->val.size()==0)
    throw std::runtime_error("Cannot convert empty list of lists to matrix");

  size_t n = force<vector_value>(r->val[0].get())->val.size();
  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,r->val.size()));
  for(size_t j=0; j<r->val.size(); ++j)
  { int_Vector col = row_to_weight(*force<row_value>(r->val[j].get()));
    if (col.size()!=n)
      throw std::runtime_error("List sizes differ in conversion to matrix");

    m->val.set_column(j,col);
  }
  push_value(std::move(m));
}

own_row weight_to_row(const int_Vector& v)
{ own_row result = std::make_shared<row_value>(v.size());
  for(size_t i=0; i<v.size(); ++i)
    result->val[i]=std::make_shared<int_value>(v[i]);
  return result;
}

void vector_intlist_convert()
{ shared_vector v = get<vector_value>();
  push_value(weight_to_row(v->val));
}

void matrix_veclist_convert()
{ shared_matrix m=get<matrix_value>();
  own_row result = std::make_shared<row_value>(m->val.numColumns());
  for(size_t i=0; i<m->val.numColumns(); ++i)
    result->val[i]=std::make_shared<vector_value>(m->val.column(i));
  push_value(std::move(result));
}

void matrix_intlistlist_convert()
{ shared_matrix m=get<matrix_value>();
  own_row result = std::make_shared<row_value>(m->val.numColumns());
  for(size_t i=0; i<m->val.numColumns(); ++i)
    result->val[i]= weight_to_row(m->val.column(i));

  push_value(std::move(result));
}

void plus_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(i+j));
}

void minus_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(i-j));
}

void times_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(i*j));
}

void divide_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (j==0) throw std::runtime_error("Division by zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>
     (j>0 ? arithmetic::divide(i,j) : -arithmetic::divide(i,-j)));
}

void modulo_wrapper(expression_base::level l)
{ int  j=get<int_value>()->val; int i=get<int_value>()->val;
  if (j==0) throw std::runtime_error("Modulo zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>
      (arithmetic::remainder(i,arithmetic::abs(j))));
}

void divmod_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (j==0) throw std::runtime_error("DivMod by zero");
  if (l!=expression_base::no_value)
  { push_value(std::make_shared<int_value>
     (j>0 ? arithmetic::divide(i,j) : -arithmetic::divide(i,-j)));
    push_value(std::make_shared<int_value>
      (arithmetic::remainder(i,arithmetic::abs(j))));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void unary_minus_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(-i));
}

void power_wrapper(expression_base::level l)
{ static shared_int one = std::make_shared<int_value>(1);
  static shared_int minus_one  = std::make_shared<int_value>(-1);
int n=get<int_value>()->val; int i=get<int_value>()->val;
  if (arithmetic::abs(i)!=1 and n<0)
    throw std::runtime_error("Negative power of integer");
  if (l==expression_base::no_value)
    return;

  if (i==1)
  { push_value(one);
      return;
  }
  if (i==-1)
  { push_value(n%2==0 ? one : minus_one);
      return;
  }

  push_value(std::make_shared<int_value>(arithmetic::power(i,n)));
}

void fraction_wrapper(expression_base::level l)
{ int d=get<int_value>()->val; int n=get<int_value>()->val;
  if (d==0) throw std::runtime_error("fraction with zero denominator");
  if (d<0) { d=-d; n=-n; } // ensure denominator is positive
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(Rational(n,d)));
}


void unfraction_wrapper(expression_base::level l)
{ Rational q=get<rat_value>()->val;
  if (l!=expression_base::no_value)
  { push_value(std::make_shared<int_value>(q.numerator()));
    push_value(std::make_shared<int_value>(q.denominator()));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void rat_plus_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val;
  Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i+j));
}

void rat_minus_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val;
  Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i-j));
}

void rat_times_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val;
  Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i*j));
}

void rat_divide_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val;
  Rational i=get<rat_value>()->val;
  if (j.numerator()==0)
    throw std::runtime_error("Rational division by zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i/j));
}

void rat_unary_minus_wrapper(expression_base::level l)
{ Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(Rational(0)-i)); }

void rat_inverse_wrapper(expression_base::level l)
{ Rational i=get<rat_value>()->val;
  if (i.numerator()==0)
    throw std::runtime_error("Inverse of zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(Rational(1)/i)); }

void rat_power_wrapper(expression_base::level l)
{ int n=get<int_value>()->val; Rational b=get<rat_value>()->val;
  if (b.numerator()==0 and n<0)
    throw std::runtime_error("Negative power of zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(b.power(n)));
}

void int_eq_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i==j));
}

void int_neq_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i!=j));
}

void int_less_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i<j));
}

void int_lesseq_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i<=j));
}

void int_greater_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i>j));
}

void int_greatereq_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i>=j));
}

void rat_eq_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val; Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i==j));
}

void rat_neq_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val; Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i!=j));
}

void rat_less_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val; Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i<j));
}

void rat_lesseq_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val; Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i<=j));
}

void rat_greater_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val; Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i>j));
}

void rat_greatereq_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val; Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i>=j));
}

void equiv_wrapper(expression_base::level l)
{ bool a=get<bool_value>()->val; bool b=get<bool_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(a==b));
}

void inequiv_wrapper(expression_base::level l)
{ bool a=get<bool_value>()->val; bool b=get<bool_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(a!=b));
}

void string_eq_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i->val==j->val));
}

void string_leq_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i->val<=j->val));
}

void concatenate_wrapper(expression_base::level l)
{ shared_string b=get<string_value>(); shared_string a=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<string_value>(a->val+b->val));
}

void int_format_wrapper(expression_base::level l)
{ int n=get<int_value>()->val;
  std::ostringstream o; o<<n;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<string_value>(o.str()));
}

void string_to_ascii_wrapper(expression_base::level l)
{ shared_string c=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>
      (c->val.size()==0 ? -1 : (unsigned char)c->val[0]));
}

void ascii_char_wrapper(expression_base::level l)
{ int c=get<int_value>()->val;
  if (c<' ' or c>'~')
    throw std::runtime_error("Value "+str(c)+" out of range");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<string_value>(std::string(1,c)));
}

void sizeof_string_wrapper(expression_base::level l)
{ size_t s=get<string_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(s));
}

void sizeof_vector_wrapper(expression_base::level l)
{ size_t s=get<vector_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(s));
}


void matrix_bounds_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  push_value(std::make_shared<int_value>(m->val.numRows()));
  push_value(std::make_shared<int_value>(m->val.numColumns()));
  if (l==expression_base::single_value)
    wrap_tuple<2>();
}

void vector_suffix_wrapper(expression_base::level l)
{ int e=get<int_value>()->val;
  own_vector r=get_own<vector_value>();
  if (l!=expression_base::no_value)
  { r->val.push_back(e);
    push_value(r);
  }
}

void vector_prefix_wrapper(expression_base::level l)
{ own_vector r=get_own<vector_value>();
  int e=get<int_value>()->val;
  if (l!=expression_base::no_value)
  { r->val.insert(r->val.begin(),e);
    push_value(r);
  }
}

void join_vectors_wrapper(expression_base::level l)
{ shared_vector y=get<vector_value>();
  shared_vector x=get<vector_value>();
  if (l!=expression_base::no_value)
  { own_vector result = std::make_shared<vector_value>(std::vector<int>());
    result->val.reserve(x->val.size()+y->val.size());
    result->val.insert(result->val.end(),x->val.begin(),x->val.end());
    result->val.insert(result->val.end(),y->val.begin(),y->val.end());
    push_value(std::move(result));
  }

}

void error_wrapper(expression_base::level l)
{ throw std::runtime_error(get<string_value>()->val); }

void vec_eq_wrapper(expression_base::level l)
{ shared_vector j=get<vector_value>(); shared_vector i=get<vector_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i->val==j->val));
}

void vec_neq_wrapper(expression_base::level l)
{ shared_vector j=get<vector_value>(); shared_vector i=get<vector_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i->val!=j->val));
}

void mat_eq_wrapper(expression_base::level l)
{ shared_matrix j=get<matrix_value>(); shared_matrix i=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i->val==j->val));
}

void mat_neq_wrapper(expression_base::level l)
{ shared_matrix j=get<matrix_value>(); shared_matrix i=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<bool_value>(i->val!=j->val));
}

void vector_div_wrapper(expression_base::level l)
{ int n=get<int_value>()->val;
  shared_vector v=get<vector_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(v->val,n));
}

void ratvec_unfraction_wrapper(expression_base::level l)
{ shared_rational_vector v = get<rational_vector_value>();
  if (l!=expression_base::no_value)
  { Weight num(v->val.numerator().begin(),v->val.numerator().end()); // convert
    push_value(std::make_shared<vector_value>(num));
    push_value(std::make_shared<int_value>(v->val.denominator()));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void ratvec_plus_wrapper(expression_base::level l)
{ shared_rational_vector v1= get<rational_vector_value>();
  shared_rational_vector v0= get<rational_vector_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(v0->val+v1->val));
}

void ratvec_minus_wrapper(expression_base::level l)
{ shared_rational_vector v1= get<rational_vector_value>();
  shared_rational_vector v0= get<rational_vector_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(v0->val-v1->val));
}

void mv_prod_wrapper(expression_base::level l)
{ shared_vector v=get<vector_value>();
  shared_matrix m=get<matrix_value>();
  if (m->val.numColumns()!=v->val.size())
    throw std::runtime_error(std::string("Size mismatch ")
     + str(m->val.numColumns()) + ":" + str(v->val.size()));
  if (l!=expression_base::no_value)
    push_value(std::make_shared<vector_value>(m->val*v->val));
}

void mrv_prod_wrapper(expression_base::level l)
{ shared_rational_vector v=get<rational_vector_value>();
  shared_matrix m=get<matrix_value>();
  if (m->val.numColumns()!=v->val.size())
    throw std::runtime_error(std::string("Size mismatch ")
     + str(m->val.numColumns()) + ":" + str(v->val.size()));
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(m->val*v->val));
}

void vv_prod_wrapper(expression_base::level l)
{ shared_vector w=get<vector_value>();
  shared_vector v=get<vector_value>();
  if (v->val.size()!=w->val.size())
    throw std::runtime_error(std::string("Size mismatch ")
     + str(v->val.size()) + ":" + str(w->val.size()));
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(v->val.dot(w->val)));
}

void vm_prod_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>(); // right factor
  shared_vector v=get<vector_value>(); // left factor
  if (v->val.size()!=m->val.numRows())
  { std::ostringstream s;
    s<< "Size mismatch " << v->val.size() << ":" << m->val.numRows();
    throw std::runtime_error(s.str());
  }
  if (l!=expression_base::no_value)
    push_value(std::make_shared<vector_value>(m->val.right_prod(v->val)));
}

void stack_rows_wrapper(expression_base::level l)
{ shared_row r = get<row_value>();
  size_t n = r->val.size();
  std::vector<const int_Vector*> row(n);
  size_t width=0; // maximal length of vectors
  for(size_t i=0; i<n; ++i)
  { row[i] = & force<vector_value>(r->val[i].get())->val;
    if (row[i]->size()>width)
      width=row[i]->size();
  }

  if (l==expression_base::no_value)
    return;

  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,width,0));
  for(size_t i=0; i<n; ++i)
    for (size_t j=0; j<row[i]->size(); ++j)
      m->val(i,j)=(*row[i])[j];
  push_value(std::move(m));
}

void combine_columns_wrapper(expression_base::level l)
{ shared_row r = get<row_value>();
  int n = get<int_value>()->val;
  if (n<0)
    throw std::runtime_error("Negative number "+str(n)+" of rows requested");

  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,r->val.size()));
  for(size_t j=0; j<r->val.size(); ++j)
  { const int_Vector& col = force<vector_value>(r->val[j].get())->val;
    if (col.size()!=size_t(n))
      throw std::runtime_error("Column "+str(j)+" size "+str(col.size())
          +" does not match specified size "+str(n));

    m->val.set_column(j,col);
  }
  if (l!=expression_base::no_value)
    push_value(std::move(m));
}

void combine_rows_wrapper(expression_base::level l)
{ shared_row r =get<row_value>();
  int n = get<int_value>()->val;
  if (n<0)
    throw std::runtime_error("Negative number "+str(n)+" of columns requested");

  own_matrix m = std::make_shared<matrix_value>(int_Matrix(r->val.size(),n));
  for(size_t i=0; i<r->val.size(); ++i)
  { const int_Vector& row = force<vector_value>(r->val[i].get())->val;
    if (row.size()!=size_t(n))
      throw std::runtime_error("Row "+str(i)+" size "+str(row.size())
          +" does not match specified size "+str(n));

    m->val.set_row(i,row);
  }
  if (l!=expression_base::no_value)
    push_value(std::move(m));
}

void section_wrapper(expression_base::level l)
{
  shared_matrix m=get<matrix_value>();
  BinaryMap A(m->val);
  BinaryMap B=A.section();
  own_matrix res = std::make_shared<matrix_value>(
    int_Matrix(B.numRows(),B.numColumns()));
  for (unsigned int j=B.numColumns(); j-->0;)
    res->val.set_column(j,int_Vector(B.column(j)));
  if (l!=expression_base::no_value)
    push_value(std::move(res));
}

void null_vec_wrapper(expression_base::level lev)
{ int l=get<int_value>()->val;
  if (lev!=expression_base::no_value)
    push_value(std::make_shared<vector_value>(int_Vector(std::abs(l),0)));
}
 void null_mat_wrapper(expression_base::level lev)
{ int l=get<int_value>()->val;
  int k=get<int_value>()->val;
  if (lev!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>
      (int_Matrix(std::abs(k),std::abs(l),0)));
}
void transpose_vec_wrapper(expression_base::level l)
{ shared_vector v=get<vector_value>();
  if (l!=expression_base::no_value)
  { own_matrix m = std::make_shared<matrix_value>(int_Matrix(1,v->val.size()));
    for (size_t j=0; j<v->val.size(); ++j)
      m->val(0,j)=v->val[j];
    push_value(std::move(m));
  }
}

void diagonal_wrapper(expression_base::level l)
{ shared_vector d=get<vector_value>();
  if (l==expression_base::no_value)
    return;
  size_t n=d->val.size();
  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n));
  for (size_t i=0; i<n; ++i)
    m->val(i,i)=d->val[i];
  push_value(std::move(m));
}

void echelon_wrapper(expression_base::level l)
{ own_matrix M=get_own<matrix_value>();
  if (l!=expression_base::no_value)
  { BitMap pivots=matreduc::column_echelon(M->val);
    push_value(M);
    own_row p_list = std::make_shared<row_value>(0);
    p_list->val.reserve(pivots.size());
    for (BitMap::iterator it=pivots.begin(); it(); ++it)
      p_list->val.push_back(std::make_shared<int_value>(*it));
    push_value(std::move(p_list));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void diagonalize_wrapper(expression_base::level l)
{ shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
  { own_matrix row = std::make_shared<matrix_value>(int_Matrix());
    own_matrix column = std::make_shared<matrix_value>(int_Matrix());
    own_vector diagonal = std::make_shared<vector_value>
       (matreduc::diagonalise(M->val,row->val,column->val));
    push_value(std::move(diagonal));
    push_value(std::move(row));
    push_value(std::move(column));
    if (l==expression_base::single_value)
      wrap_tuple<3>();
  }
}

void adapted_basis_wrapper(expression_base::level l)
{ shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
  { own_vector diagonal = std::make_shared<vector_value>(std::vector<int>());
    push_value(std::make_shared<matrix_value>
      (matreduc::adapted_basis(M->val,diagonal->val)));
    push_value(std::move(diagonal));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void kernel_wrapper(expression_base::level l)
{ shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>(lattice::kernel(M->val)));
}

void eigen_lattice_wrapper(expression_base::level l)
{ int eigen_value = get<int_value>()->val;
  shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>
      (lattice::eigen_lattice(M->val,eigen_value)));
}

void row_saturate_wrapper(expression_base::level l)
{ shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>(lattice::row_saturate(M->val)));
}

void invfact_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  own_vector inv_factors = std::make_shared<vector_value>(std::vector<int>());
matreduc::Smith_basis(m->val,inv_factors->val);
  push_value(std::move(inv_factors));
}

void Smith_basis_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  own_vector inv_factors = std::make_shared<vector_value>(std::vector<int>());
push_value(std::make_shared<matrix_value>
    (matreduc::Smith_basis(m->val,inv_factors->val)));
}

void Smith_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  own_vector inv_factors = std::make_shared<vector_value>(std::vector<int>());
push_value(std::make_shared<matrix_value>
    (matreduc::Smith_basis(m->val,inv_factors->val)));
  push_value(std::move(inv_factors));
  if (l==expression_base::single_value)
    wrap_tuple<2>();
}

void invert_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (m->val.numRows()!=m->val.numColumns())
  { std::ostringstream s;
    s<< "Cannot invert a " 
     << m->val.numRows() << "x" << m->val.numColumns() << " matrix";
    throw std::runtime_error(s.str());
  }
  if (l==expression_base::no_value)
    return;
  own_int denom = std::make_shared<int_value>(0);
push_value(std::make_shared<matrix_value>(m->val.inverse(denom->val)));
  push_value(std::move(denom));
  if (l==expression_base::single_value)
    wrap_tuple<2>();
}

void subspace_normal_wrapper(expression_base::level l)
{
  typedef BitVector<64> bitvec;
  shared_row generators=get<row_value>();
  unsigned int n_gens = generators->val.size();
  unsigned int rank = n_gens=0 ? 0 :
    force<vector_value>(generators->val[0].get())->val.size();
  if (rank>64)
    throw std::runtime_error("Rank too large to handle "+str(rank));
  if (n_gens>64)
    throw std::runtime_error ("Too many generators to handle "+str(n_gens));

  std::vector<bitvec> basis, combination;
  std::vector<unsigned int> pivot; // |pivot[i]| is bit position for |basis[i]|
  std::vector<unsigned int> pivoter;
    // generator |pivoter[i]| led to |basis[i]|
  basis.reserve(rank); pivot.reserve(rank); pivoter.reserve(rank);
  bitvector::initBasis(combination,n_gens);

  
  for (unsigned int i=0; i<n_gens; ++i)
  { const vector_value* p=force<vector_value>(generators->val[i].get());
    if (p->val.size()!=rank)
      throw std::runtime_error
            ("Generator vector of wrong size "+str(p->val.size()));
    bitvec v(p->val); // reduce modulo $2$
    for (unsigned int j=0; j<basis.size(); ++j)
      if (v[pivot[j]])
      {
         v -= basis[j];
         combination[i] -= combination[pivoter[j]];
      }
    if (v.nonZero())
    {
      unsigned int piv = v.firstBit(); // new pivot
      for (unsigned int j=0; j<basis.size(); ++j)
      if (basis[j][piv])
      {
         basis[j] -= v;
         combination[pivoter[j]] -= combination[i];
      }
      basis.push_back(v);
      pivoter.push_back(i);
      pivot.push_back(piv);
    }
  }
  if (l!=expression_base::no_value)
  
  { Permutation pi = permutations::standardization(pivot,n_gens);
    own_row basis_r = std::make_shared<row_value>(basis.size());
    own_row combin_r = std::make_shared<row_value>(basis.size());
    own_row relations = std::make_shared<row_value>(n_gens-basis.size());
    own_row pivot_r = std::make_shared<row_value>(basis.size());
    unsigned int k=0;
    for (unsigned int i=0; i<n_gens; ++i)
      if (k<basis.size() and i==pivoter[k])
      { unsigned d = pi[k]; // destination position
        { own_vector vv = std::make_shared<vector_value>(int_Vector(rank,0));
          basis_r->val[d]=vv; // store |vv| converted to |shared_value|
          int_Vector& v = vv->val; // then use underlying vector to set the value
          for (bitvec::base_set::iterator it=basis[k].data().begin(); it(); ++it)
            v[*it]=1;
        }
        { own_vector vv = std::make_shared<vector_value>(int_Vector(n_gens,0));
          combin_r->val[d]=vv;
          int_Vector& v = vv->val;
          for (bitvec::base_set::iterator
               it=combination[i].data().begin(); it(); ++it)
            v[*it]=1;
        }
        pivot_r->val[d] = std::make_shared<int_value>(pivot[k]);
        ++k;
      }
      else
      { own_vector vv = std::make_shared<vector_value>(int_Vector(n_gens,0));
        relations->val[i-k] = vv;
        int_Vector& v = vv->val;
        for (bitvec::base_set::iterator
             it=combination[i].data().begin(); it(); ++it)
        v[*it]=1;
      }
    assert (k==basis.size());
  push_value(std::move(basis_r));
    push_value(std::move(combin_r));
    push_value(std::move(relations));
    push_value(std::move(pivot_r));
    if (l==expression_base::single_value)
      wrap_tuple<4>();
  }
}

}

void Id_table::add(id_type id, shared_value val, type_expr&& type)
{ auto its = table.equal_range(id);

  if (its.first==its.second) // no global identifier was previously known
    table.insert // better: |emplace_hint(its.first,id,@[...@])| with gcc 4.8
      (its.first,std::make_pair(id,
        id_data( std::make_shared<shared_value>(val), std::move(type) )));
  else // a global identifier was previously known
    its.first->second =
      id_data( std::make_shared<shared_value>(val), std::move(type) );
}

bool Id_table::remove(id_type id)
{ map_type::iterator p = table.find(id);
  if (p==table.end())
    return false;
  table.erase(p); return true;
}

const_type_p Id_table::type_of(id_type id) const
{ map_type::const_iterator p=table.find(id);
  return p==table.end() ? nullptr : &p->second.type();
}
void Id_table::specialise(id_type id,const type_expr& type)
{ map_type::iterator p=table.find(id);
  p->second.type().specialise(type);
}

shared_value Id_table::value_of(id_type id) const
{ map_type::const_iterator p=table.find(id);
  return p==table.end() ? shared_value(value(nullptr)) : *p->second.value();
}
shared_share Id_table::address_of(id_type id)
{ map_type::iterator p=table.find(id);
  if (p==table.end())
    throw std::logic_error 
    (std::string("Identifier without table entry:")
     +main_hash_table->name_of(id));

  return p->second.value();
}

void Id_table::print(std::ostream& out) const
{ for (map_type::const_iterator p=table.begin(); p!=table.end(); ++p)
  { out << main_hash_table->name_of(p->first) << ": " 
        << p->second.type() << ": ";
    if (*p->second.value()==nullptr)
      out << '*';
    else
      out << **p->second.value();
    out << std::endl;
   }
}

std::ostream& operator<< (std::ostream& out, const Id_table& p)
{ p.print(out); return out; }

const overload_table::variant_list& overload_table::variants
  (id_type id) const
{ static const variant_list empty;
  auto p=table.find(id);
  return p==table.end() ? empty : p->second;
}

void overload_table::add
  (id_type id, shared_value val, type_ptr tp)
{ assert (tp->kind==function_type);
  func_type type(std::move(*tp->func)); // steal the function type
  auto its = table.equal_range(id);
  if (its.first==its.second) // a fresh overloaded identifier
  { its.first=table.insert // better: |emplace_hint| with gcc 4.8
      (its.first,std::make_pair(id, variant_list()));
    its.first->second.push_back(
      overload_data( std::move(val), std::move(type)) );
  }
  else
  { variant_list& slot=its.first->second; // vector of all variants
    
    { size_t lwb=0; size_t upb=slot.size();
      for (size_t i=0; i<slot.size(); ++i)
      { unsigned int cmp= is_close(type.arg_type,slot[i].type().arg_type);
        switch (cmp)
        {
          case 0x6: lwb=i+1; break;
            // existent type |i| converts to |type|, which must come later
          case 0x5:  if (upb>i) upb=i;  break;
            // |type| converts to type |i|, so it must come before
          case 0x7: // mutually convertible types, maybe identical ones
            if (slot[i].type().arg_type==type.arg_type)
              // identical ones: overload redefinition case
            { slot[i] = overload_data(std::move(val),std::move(type));
                return;
              }
          // |else| {\bf fall through}
          case 0x4:
             
             { std::ostringstream o;
               o << "Cannot overload `" << main_hash_table->name_of(id) << "', " 
                    "previous type " << slot[i].type().arg_type
                 << " is too close to " << type.arg_type
                 << ",\nmaking overloading potentially ambiguous." 
                    " Broadness cannot disambiguate,\nas "
                 << (cmp==0x4 ? "neither" :"either") 
                 << " type converts to the other";
               throw program_error(o.str());
             }
          default: {} // nothing for unrelated argument types
        }
      }
      
      if (lwb>upb)
        throw std::logic_error("Conflicting order of related overload types");
      else
      { slot.insert // better use |emplace| with gcc 4.8
        (slot.begin()+upb,overload_data(std::move(val),std::move(type)));
      }
    }
  }
}

bool overload_table::remove(id_type id, const type_expr& arg_t)
{ map_type::iterator p=table.find(id);
  if (p==table.end()) return false; // |id| was not known at all
  variant_list& variants=p->second;
  for (size_t i=0; i<variants.size(); ++i)
    if (variants[i].type().arg_type==arg_t)
    {
      variants.erase(variants.begin()+i);
      return true;
    }
  return false; // |id| was known, but no such overload is present
}

void overload_table::print(std::ostream& out) const
{ for (auto p=table.begin(); p!=table.end(); ++p)
    for (auto it=p->second.begin(); it!=p->second.end(); ++it)
      out << main_hash_table->name_of(p->first) << ": " 
        << it->type() << ": " << *it->value() << std::endl;
}

std::ostream& operator<< (std::ostream& out, const overload_table& p)
{ p.print(out); return out; }

void initialise_evaluator()
{ 
  execution_stack.reserve(16); 
  
  coercion(int_type,rat_type, "QI", rational_convert); 
  coercion(row_of_rat_type,ratvec_type, "Qv[Q]", ratlist_ratvec_convert); 
  coercion(ratvec_type,row_of_rat_type, "[Q]Qv", ratvec_ratlist_convert); 
  coercion(vec_type,ratvec_type,"QvV", vec_ratvec_convert); 
  coercion(row_of_int_type,ratvec_type,"Qv[I]", intlist_ratvec_convert);
  
  coercion(row_of_int_type, vec_type, "V[I]", intlist_vector_convert); 
  coercion(row_of_vec_type,mat_type, "M[V]", veclist_matrix_convert); 
  coercion(row_row_of_int_type,mat_type, "M[[I]]", intlistlist_matrix_convert); 
  coercion(vec_type,row_of_int_type, "[I]V", vector_intlist_convert); 
  coercion(mat_type,row_of_vec_type, "[V]M", matrix_veclist_convert); 
  coercion(mat_type,row_row_of_int_type, "[[I]]M", matrix_intlistlist_convert); 
  
  install_function(plus_wrapper,"+","(int,int->int)");
  install_function(minus_wrapper,"-","(int,int->int)");
  install_function(times_wrapper,"*","(int,int->int)");
  install_function(divide_wrapper,"\\","(int,int->int)");
  install_function(modulo_wrapper,"%","(int,int->int)");
  install_function(divmod_wrapper,"\\%","(int,int->int,int)");
  install_function(unary_minus_wrapper,"-","(int->int)");
  install_function(power_wrapper,"^","(int,int->int)");
  install_function(fraction_wrapper,"/","(int,int->rat)");
  install_function(unfraction_wrapper,"%","(rat->int,int)");
     // unary \% means ``break open''
  install_function(rat_plus_wrapper,"+","(rat,rat->rat)");
  install_function(rat_minus_wrapper,"-","(rat,rat->rat)");
  install_function(rat_times_wrapper,"*","(rat,rat->rat)");
  install_function(rat_divide_wrapper,"/","(rat,rat->rat)");
  install_function(rat_unary_minus_wrapper,"-","(rat->rat)");
  install_function(rat_inverse_wrapper,"/","(rat->rat)");
  install_function(rat_power_wrapper,"^","(rat,int->rat)");
  install_function(int_eq_wrapper,"=","(int,int->bool)");
  install_function(int_neq_wrapper,"!=","(int,int->bool)");
  install_function(int_less_wrapper,"<","(int,int->bool)");
  install_function(int_lesseq_wrapper,"<=","(int,int->bool)");
  install_function(int_greater_wrapper,">","(int,int->bool)");
  install_function(int_greatereq_wrapper,">=","(int,int->bool)");
  install_function(rat_eq_wrapper,"=","(rat,rat->bool)");
  install_function(rat_neq_wrapper,"!=","(rat,rat->bool)");
  install_function(rat_less_wrapper,"<","(rat,rat->bool)");
  install_function(rat_lesseq_wrapper,"<=","(rat,rat->bool)");
  install_function(rat_greater_wrapper,">","(rat,rat->bool)");
  install_function(rat_greatereq_wrapper,">=","(rat,rat->bool)");
  install_function(equiv_wrapper,"=","(bool,bool->bool)");
  install_function(inequiv_wrapper,"!=","(bool,bool->bool)");
  install_function(string_eq_wrapper,"=","(string,string->bool)");
  install_function(string_leq_wrapper,"<=","(string,string->bool)");
  install_function(concatenate_wrapper,"#","(string,string->string)");
  install_function(int_format_wrapper,"int_format","(int->string)");
  install_function(string_to_ascii_wrapper,"ascii","(string->int)");
  install_function(ascii_char_wrapper,"ascii","(int->string)");
  install_function(sizeof_string_wrapper,"#","(string->int)");
  install_function(sizeof_vector_wrapper,"#","(vec->int)");
  install_function(matrix_bounds_wrapper,"#","(mat->int,int)");
  install_function(vector_div_wrapper,"/","(vec,int->ratvec)");
  install_function(ratvec_unfraction_wrapper,"%","(ratvec->vec,int)");
  install_function(ratvec_plus_wrapper,"+","(ratvec,ratvec->ratvec)");
  install_function(ratvec_minus_wrapper,"-","(ratvec,ratvec->ratvec)");
  install_function(error_wrapper,"error","(string->*)");
  install_function(vector_suffix_wrapper,"#","(vec,int->vec)");
  install_function(vector_prefix_wrapper,"#","(int,vec->vec)");
  install_function(join_vectors_wrapper,"#","(vec,vec->vec)");
  install_function(vec_eq_wrapper,"=","(vec,vec->bool)");
  install_function(vec_neq_wrapper,"!=","(vec,vec->bool)");
  install_function(mat_eq_wrapper,"=","(mat,mat->bool)");
  install_function(mat_neq_wrapper,"!=","(mat,mat->bool)");
  install_function(vv_prod_wrapper,"*","(vec,vec->int)");
  install_function(mrv_prod_wrapper,"*","(mat,ratvec->ratvec)");
  install_function(mv_prod_wrapper,"*","(mat,vec->vec)");
  install_function(mm_prod_wrapper,"*","(mat,mat->mat)");
  install_function(vm_prod_wrapper,"*","(vec,mat->vec)");
  install_function(stack_rows_wrapper,"stack_rows","([vec]->mat)");
  install_function(combine_columns_wrapper,"#","(int,[vec]->mat)");
  install_function(combine_rows_wrapper,"^","(int,[vec]->mat)");
  install_function(section_wrapper,"mod2_inverse","(mat->mat)");
  
  install_function(null_vec_wrapper,"null","(int->vec)");
  install_function(null_mat_wrapper,"null","(int,int->mat)");
  install_function(transpose_vec_wrapper,"^","(vec->mat)");
  install_function(transpose_mat_wrapper,"^","(mat->mat)");
  install_function(id_mat_wrapper,"id_mat","(int->mat)");
  install_function(diagonal_wrapper,"diagonal","(vec->mat)");
  install_function(echelon_wrapper,"echelon","(mat->mat,[int])");
  install_function(diagonalize_wrapper,"diagonalize","(mat->vec,mat,mat)");
  install_function(adapted_basis_wrapper,"adapted_basis","(mat->mat,vec)");
  install_function(kernel_wrapper,"kernel","(mat->mat)");
  install_function(eigen_lattice_wrapper,"eigen_lattice","(mat,int->mat)");
  install_function(row_saturate_wrapper,"row_saturate","(mat->mat)");
  install_function(invfact_wrapper,"inv_fact","(mat->vec)");
  install_function(Smith_basis_wrapper,"Smith_basis","(mat->mat)");
  install_function(Smith_wrapper,"Smith","(mat->mat,vec)");
  install_function(invert_wrapper,"invert","(mat->mat,int)");
  install_function(subspace_normal_wrapper,
     "subspace_normal","([vec]->[vec],[vec],[vec],[int])");
 }

type_expr analyse_types(const expr& e,expression_ptr& p)
  throw(std::bad_alloc,std::runtime_error)
{ try
  { type_expr type=unknown_type.copy();
    p = convert_expr(e,type);
    return type;
  }
  catch (type_error& err)
  { std::cerr << err.what() << ":\n  Subexpression " << err.offender << 
    " has wrong type: found " << err.actual << 
    " while " << err.required << " was needed.\n";

  }
  catch (expr_error& err)
  { std::cerr << "Error in expression " << err.offender << "\n  "
              << err.what() << std::endl;
  }
  catch (program_error& err)
  { std::cerr << "Error during analysis of expression " << e << "\n  "
                << err.what() << std::endl;
  }
  throw std::runtime_error("Type check failed");

}

void global_set_identifier(id_pat pat, expr_p raw, int overload)
{ expr_ptr saf(raw); const expr& rhs(*raw);
  size_t n_id=count_identifiers(pat);
  static const char* phase_name[3] = {"type_check","evaluation","definition"};
  int phase=0; // needs to be declared outside the |try|, is used in |catch|
  try
  { expression_ptr e;
    type_expr t=analyse_types(rhs,e);
    if (not pattern_type(pat).specialise(t))
      
      { std::ostringstream o;
        o << "Type " << t 
          << " of right hand side does not match required pattern "
          << pattern_type(pat);
        throw std::runtime_error(o.str());
      }
    if (overload!=0)
      
      { bool clear = t.kind!=function_type;
          // cannot overload with a non-function value
        if (not clear)
        { type_expr& arg=t.func->arg_type;
        clear = arg.kind==tuple_type and arg.tupple.empty();
           // nor parameterless functions
        }
        if (clear and overload==2) // inappropriate function type with operator
        { std::string which(t.kind==function_type ? "parameterless " : "non-");
          throw std::runtime_error
            ("Cannot set operator to a "+which+"function value");
        }
        if (clear)
          overload=0;
      }

    phase=1;
    layer b(n_id);
    thread_bindings(pat,t,b); // match identifiers and their future types

    std::vector<shared_value> v;
    v.reserve(n_id);
  e->eval();
    thread_components(pat,pop_value(),std::back_inserter(v));
     // associate values with identifiers

    phase=2;
    
    { unsigned int input_level = main_input_buffer->include_depth();
      std::cout << std::setw(2*input_level) << "";
    }
    if (overload==0)
      
      { if (n_id>0)
          std::cout << "Identifier";
        for (size_t i=0; i<n_id; ++i)
        { std::cout << (i==0 ? n_id==1 ? " " : "s " : ", ") 
                    << main_hash_table->name_of(b[i].first);
          if (global_id_table->present(b[i].first))
            std::cout << " (overriding previous)";
          std::cout << ": " << b[i].second;
          global_id_table->add(b[i].first,v[i],b[i].second.copy());
        }
      }
    else
      
      { assert(n_id=1);
        size_t old_n=global_overload_table->variants(b[0].first).size();
        global_overload_table->add(b[0].first,v[0],acquire(&b[0].second));
          // insert or replace table entry
        size_t n=global_overload_table->variants(b[0].first).size();
        if (n==old_n)
          std::cout << "Redefined ";
        else if (n==1)
          std::cout << "Defined ";
        else
          std::cout << "Added definition [" << n << "] of ";
        std::cout << main_hash_table->name_of(b[0].first) << ": " << b[0].second;
      }
    std::cout << std::endl;
  }
  
  catch (std::runtime_error& err)
  { std::cerr << err.what() << '\n';
    if (n_id>0)
    { std::vector<id_type> names; names.reserve(n_id);
      list_identifiers(pat,names);
      std::cerr << "  Identifier" << (n_id==1 ? "" : "s");
      for (size_t i=0; i<n_id; ++i)
        std::cerr << (i==0 ? " " : ", ") << main_hash_table->name_of(names[i]);
      std::cerr << " not " << (overload==0 ? "created." : "overloaded.")
                << std::endl;
    }
    reset_evaluator(); main_input_buffer->close_includes();
  }
  catch (std::logic_error& err)
  { std::cerr << "Unexpected error: " << err.what() << ", " 
              << phase_name[phase]
              << " aborted.\n";
  reset_evaluator(); main_input_buffer->close_includes();
  }
  catch (std::exception& err)
  { std::cerr << err.what() << ", "
              << phase_name[phase]
              << " aborted.\n";
  reset_evaluator(); main_input_buffer->close_includes();
  }
}

void global_declare_identifier(id_type id, type_p t)
{ value undef=nullptr;
  global_id_table->add(id,shared_value(undef),t->copy());
  
  { unsigned int input_level = main_input_buffer->include_depth();
    std::cout << std::setw(2*input_level) << "";
  }
  std::cout << "Identifier " << main_hash_table->name_of(id)
            << " : " << *t << std::endl;
}

void global_forget_identifier(id_type id)
{ std::cout << "Identifier " << main_hash_table->name_of(id)
            << (global_id_table->remove(id) ? " forgotten" : " not known")
            << std::endl;
}

void global_forget_overload(id_type id, type_p t)
{ const type_expr& type=*t;
  std::cout << "Definition of " << main_hash_table->name_of(id)
            << '@' << type 
            << (global_overload_table->remove(id,type)
               ? " forgotten"
               : " not known")
            << std::endl;
}

void type_of_expr(expr_p raw)
{ expr_ptr saf(raw); const expr& e=*raw;
  try
  { expression_ptr p;
    *output_stream << "type: " << analyse_types(e,p) << std::endl;
  }
  catch (std::exception& err) { std::cerr<<err.what()<<std::endl; }
}

void show_overloads(id_type id)
{ const overload_table::variant_list& variants =
   global_overload_table->variants(id);
  *output_stream
   << (variants.empty() ? "No overloads for " : "Overloaded instances of ") 
   << main_hash_table->name_of(id) << std::endl;
 for (size_t i=0; i<variants.size(); ++i)
   *output_stream << "  "
    << variants[i].type().arg_type << "->" << variants[i].type().result_type 
    << std::endl;
}

void show_ids()
{ *output_stream << "Overloaded operators and functions:\n"
                 << *global_overload_table 
                 << "Global variables:\n" << *global_id_table;
}

void vector_value::print(std::ostream& out) const
{ size_t l=val.size(),w=0; std::vector<std::string> tmp(l);
  for (size_t i=0; i<l; ++i)
  { std::ostringstream s; s<<val[i]; tmp[i]=s.str();
    if (tmp[i].length()>w) w=tmp[i].length();
  }
  if (l==0) out << "[ ]";
  else
  { w+=1; out << std::right << '[';
    for (size_t i=0; i<l; ++i)
      out << std::setw(w) << tmp[i] << (i<l-1 ? "," : " ]");
  }
}

void rational_vector_value::print(std::ostream& out) const
{ size_t l=val.size(),w=0; std::vector<std::string> tmp(l);
  for (size_t i=0; i<l; ++i)
  { std::ostringstream s; s<<val.numerator()[i]; tmp[i]=s.str();
    if (tmp[i].length()>w) w=tmp[i].length();
  }
  if (l==0) out << "[ ]";
  else
  { w+=1; out << std::right << '[';
    for (size_t i=0; i<l; ++i)
      out << std::setw(w) << tmp[i] << (i<l-1 ? "," : " ]");
  }
  out << '/' << val.denominator();
}

void matrix_value::print(std::ostream& out) const
{ size_t k=val.numRows(),l=val.numColumns();
  if (k==0 or l==0)
  {  out << "The " << k << 'x' << l << " matrix"; return; }
  std::vector<size_t> w(l,0);
  for (size_t i=0; i<k; ++i)
    for (size_t j=0; j<l; ++j)
    { std::ostringstream s; s<<val(i,j); size_t len=s.str().length();
      if (len>w[j]) w[j]=len;
    }
  out << std::endl << std::right;
  for (size_t i=0; i<k; ++i)
  { out << '|';
    for (size_t j=0; j<l; ++j)
      out << std::setw(w[j]+1) << val(i,j) << (j<l-1 ? ',' : ' ');
    out << '|' << std::endl;
  }
}

void install_function
 (wrapper_function f,const char*name, const char* type_string)
{ type_ptr type = mk_type(type_string);
  std::ostringstream print_name; print_name<<name;
  if (type->kind!=function_type)
    throw std::logic_error
     ("Built-in with non-function type: "+print_name.str());
  if (type->func->arg_type==void_type)
  { own_value val = std::make_shared<builtin_value>(f,print_name.str());
    global_id_table->add
      (main_hash_table->match_literal(name),val,std::move(*type));
  }
  else
  { print_name << '@' << type->func->arg_type;
    own_value val = std::make_shared<builtin_value>(f,print_name.str());
    global_overload_table->add
      (main_hash_table->match_literal(name),val,std::move(type));
  }
}

void mm_prod_wrapper(expression_base::level l)
{ shared_matrix rf=get<matrix_value>(); // right factor
  shared_matrix lf=get<matrix_value>(); // left factor
  if (lf->val.numColumns()!=rf->val.numRows())
  { std::ostringstream s;
    s<< "Size mismatch " << lf->val.numColumns() << ":" << rf->val.numRows();
    throw std::runtime_error(s.str());
  }
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>(lf->val*rf->val));
}

 void transpose_mat_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>(m->val.transposed()));
}

void id_mat_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>
      (int_Matrix(std::abs(i)))); // identity
}

} }

