
#ifndef GLOBAL_H
#define GLOBAL_H


#include <map>
#include "lexer.h" 

#include "types.h"
#include "parsetree.h"

#include <string>
#include <sstream>

#include "arithmetic.h"

#include "atlas_types.h" // type declarations that are ``common knowledge''
#include "types.h" // base types we derive from
#include "matrix.h" // to make |int_Vector| and |int_Matrix| complete types
#include "ratvec.h" 

namespace atlas { namespace interpreter {

typedef std::shared_ptr<shared_value> shared_share;
class id_data
{ shared_share val;  type_expr tp;
public:
  id_data(shared_share&& val,type_expr&& t)
  : val(std::move(val)), tp(std::move(t)) {}
  id_data (id_data&& x) = default;
  void swap(id_data& x)  { val.swap(x.val); tp.swap(x.tp); }
  id_data& operator=(id_data&& x) = default; // no copy-and-swap needed

  shared_share value() const { return val; }
  const type_expr& type() const { return tp; }
  type_expr& type() { return tp; }
  // non-|const| reference; may be specialised by caller
};

class Id_table
{ typedef std::map<id_type,id_data> map_type;
  map_type table;
public:
  Id_table(const Id_table&) =  delete ;
  Id_table& operator=(const Id_table&) =  delete ;
  Id_table() : table() {} // the default and only constructor

  void add(id_type id, shared_value v, type_expr&& t); // insertion
  bool remove(id_type id); // deletion
  shared_share address_of(id_type id); // locate

  bool present (id_type id) const
  { return table.find(id)!=table.end(); }
  const_type_p type_of(id_type id) const;
  // pure lookup, may return |nullptr|
  void specialise(id_type id,const type_expr& type);
  // specialise type stored for identifier
  shared_value value_of(id_type id) const; // look up

  size_t size() const { return table.size(); }
  void print(std::ostream&) const;
};

struct overload_data
{ shared_value val;  func_type tp;
public:
  overload_data(shared_value&& val,func_type&& t)
  : val(std::move(val)), tp(std::move(t)) {}
  overload_data (overload_data&& x) = default;
  overload_data& operator=(overload_data&& x) = default; // no copy-and-swap needed

  shared_value value() const { return val; }
  const func_type& type() const { return tp; }
};

class overload_table
{
public:
  typedef std::vector<overload_data> variant_list;
  typedef std::map<id_type,variant_list> map_type;
private:
  map_type table;
public:
  overload_table (const Id_table&) =delete;
  overload_table& operator=(const Id_table&) = delete;
  overload_table() : table() {} // the default and only constructor
 // accessors
  const variant_list& variants(id_type id) const;
  size_t size() const { return table.size(); }
   // number of distinct identifiers
  void print(std::ostream&) const;
 // manipulators
  void add(id_type id, shared_value v, type_ptr t);
   // insertion
  bool remove(id_type id, const type_expr& arg_t); //deletion
};

struct int_value : public value_base
{ int val;

  explicit int_value(int v) : val(v)  {}
  ~int_value() {}
  void print(std::ostream& out) const { out << val; }
  int_value* clone() const { return new int_value(*this); }
  static const char* name() { return "integer"; }
private:
  int_value(const int_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const int_value> shared_int;
typedef std::shared_ptr<int_value> own_int;

struct rat_value : public value_base
{ Rational val;

  explicit rat_value(Rational v) : val(v)  {}
  ~rat_value() {}
  void print(std::ostream& out) const { out << val; }
  rat_value* clone() const { return new rat_value(*this); }
  static const char* name() { return "integer"; }
private:
  rat_value(const rat_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const rat_value> shared_rat;

struct string_value : public value_base
{ std::string val;

  explicit string_value(const std::string& s) : val(s)  {}
  ~string_value() {}
  void print(std::ostream& out) const { out << '"' << val << '"'; }
  string_value* clone() const { return new string_value(*this); }
  static const char* name() { return "string"; }
private:
  string_value(const string_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const string_value> shared_string;


struct bool_value : public value_base
{ bool val;

  explicit bool_value(bool v) : val(v)  {}
  ~bool_value() {}
  void print(std::ostream& out) const { out << std::boolalpha << val; }
  bool_value* clone() const { return new bool_value(*this); }
  static const char* name() { return "Boolean"; }
private:
  bool_value(const bool_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const bool_value> shared_bool;

struct vector_value : public value_base
{ int_Vector val;

  explicit vector_value(const std::vector<int>& v) : val(v)  {}
  ~vector_value() {}
  virtual void print(std::ostream& out) const;
  vector_value* clone() const { return new vector_value(*this); }
  static const char* name() { return "vector"; }
private:
  vector_value(const vector_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const vector_value> shared_vector;
typedef std::shared_ptr<vector_value> own_vector;

struct matrix_value : public value_base
{ int_Matrix val;

  explicit matrix_value(const int_Matrix& v) : val(v)  {}
  ~matrix_value() {}
  virtual void print(std::ostream& out) const;
  matrix_value* clone() const { return new matrix_value(*this); }
  static const char* name() { return "matrix"; }
private:
  matrix_value(const matrix_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const matrix_value> shared_matrix;
typedef std::shared_ptr<matrix_value> own_matrix;

struct rational_vector_value : public value_base
{ RatWeight val;

  explicit rational_vector_value(const RatWeight& v)
   : val(v){ val.normalize();}
  rational_vector_value(const int_Vector& v,int d)
   : val(v,d)  { val.normalize(); }
  ~rational_vector_value() {}
  virtual void print(std::ostream& out) const;
  rational_vector_value* clone() const
   { return new rational_vector_value(*this); }
  static const char* name() { return "rational vector"; }
private:
  rational_vector_value(const rational_vector_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const rational_vector_value> shared_rational_vector;

typedef void (* wrapper_function)(expression_base::level);


extern Id_table* global_id_table;

extern overload_table* global_overload_table;

extern std::ostream* output_stream;

extern int verbosity;


std::ostream& operator<< (std::ostream& out, const Id_table& p);

std::ostream& operator<< (std::ostream& out, const overload_table& p);

void initialise_evaluator();

type_expr analyse_types(const expr& e,expression_ptr& p)
   throw(std::bad_alloc,std::runtime_error);

void global_set_identifier(struct id_pat id, expr_p e, int overload);
void global_declare_identifier(id_type id, type_p type);
void global_forget_identifier(id_type id);
void global_forget_overload(id_type id, type_p type);
void show_ids();
void type_of_expr(expr_p e);
void show_overloads(id_type id);

void install_function
 (wrapper_function f,const char*name, const char* type_string);

void mm_prod_wrapper (expression_base::level);

void transpose_mat_wrapper (expression_base::level);
void id_mat_wrapper(expression_base::level l);


template <typename T>
  std::string str(T n) { std::ostringstream s; s<<n; return s.str(); }
inline std::string str(unsigned char c)
  { return str(static_cast<unsigned int>(c)); }

} }
#endif

