
#include "lexer.h"
using namespace std; 
namespace atlas
{ namespace interpreter
   {

Lexical_analyser::Lexical_analyser
  (BufferedInput& source, Hash_table& hash,
   const char** keywords, const char** type_names)
: input(source),id_table(hash),nesting(0)
 ,prevent_termination('\0'),state(initial)
{ 
  { for (size_t i=0; keywords[i]!=0; ++i)
      id_table.match_literal(keywords[i]);
    keyword_limit=id_table.nr_entries();
    for (size_t i=0; type_names[i]!=0; ++i)
      id_table.match_literal(type_names[i]);
    type_limit=id_table.nr_entries();
  }
  comment_start=comment_end=256; // a non-|char| value
}

bool Lexical_analyser::reset()
{ nesting=0; state=initial; input.reset(); return input.getline(); }

void Lexical_analyser::skip_space(void)
{ if (prevent_termination!='\0') input.push_prompt(prevent_termination);
  do
  { char c=input.shift();
    if (std::isspace(c))
    { if (c=='\n' and prevent_termination=='\0' and nesting==0
          or c=='\f')
        break;
// newline not skipped if some command could end here, form feed never skipped
      continue; // all other space is skipped
    }
    if (c==comment_start)
      if (comment_end=='\n' or comment_end==comment_start)
      { do c=input.shift(); while (c!=comment_end && c!='\0');
        input.unshift(); // reconsider newline character
      }
      else // skip comment, allowing for nested comment groups
      { int level=1; int line; int column;
        input.locate(input.point(),line,column); // maybe needed for reporting
        input.push_prompt(comment_start);
        do
        {
          do c=input.shift();
          while (c!= comment_start and c!=comment_end and c!='\0' and c!='\f');
          if (c=='\0' or c=='\f')
          { std::cerr << "Comment that started on line " << line
                      << ", column " << column  << " is never closed.\n";
           input.unshift(); break;
          // force out of comment loop, reconsider character
          }
          if (c==comment_start)
          { ++level;
            input.push_prompt(comment_start);
          }
          else
          { --level;
            input.pop_prompt();
          } // |c==comment_end|
        }
        while (level>0);
      }
    else break; // non-space and non-comment character
  } while(true);
  if (prevent_termination!='\0') input.pop_prompt();
  input.unshift(); // prepare to re-read character that ended space
}

std::string Lexical_analyser::scan_quoted_string()
{ const char* start=input.point(); bool broken=false;
  std::string result; char c;
  while (true)
  { for (c=input.shift(); c!='"' and c!='\n' and c!='\0'; c=input.shift())
      result.push_back(c);
    if (c!='"')
      { broken=true; break; }
    else if ((c=input.shift())!='"')
      break; // normal ending of string
    result.push_back(c); // doubled quote; insert one copy and continue
  }
  input.unshift();
  if (broken)
  { const char* end=input.point();
    int l0,c0,l1,c1;
    input.locate(start,l0,c0); input.locate(end,l1,c1);
    input.show_range(cerr,l0,c0,l1,c1);
    cerr << "Closing string denotation.\n";
  }
  return result;
}

int Lexical_analyser::get_token(YYSTYPE *valp, YYLTYPE* locp)
{ if (state==ended)
    { state=initial; return 0; } // send end of input
  skip_space(); prevent_termination='\0';
  input.locate(input.point(),locp->first_line,locp->first_column);
  int code; char c=input.shift();
  if (isalpha(c) or c=='_')
    
    { const char* p=input.point()-1; // start of token
      do
        c=input.shift();
      while(isalpha(c) || isdigit(c) || c=='_');
      input.unshift();
      id_type id_code=id_table.match(p,input.point()-p);
      if (id_code>=type_limit)
      { valp->id_code=id_code; code=IDENT; }
      else if (id_code>=keyword_limit)
      { valp->type_code=id_code-keyword_limit; code=TYPE; }
      else
      { code=QUIT+id_code;
        switch(code)
        {
          case LET: ++nesting; input.push_prompt('L'); break;
          case BEGIN:
          case IF:
          case WHILE:
          case FOR: ++nesting; input.push_prompt('G'); break;
          case IN: if (input.top_prompt()=='L')
          { --nesting; input.pop_prompt(); prevent_termination='I'; }
          break;
          case END:
          case FI:
          case OD: --nesting; input.pop_prompt(); break;
          case AND: case OR: case NOT: prevent_termination='~'; break;
          case WHATTYPE: prevent_termination='W'; break;
        }
      }
    }
  else if (isdigit(c))
    
    { const char* p=input.point()-1; // start of token
      do
        c=input.shift();
      while(isdigit(c));
      input.unshift();
      const char* end=input.point();
      unsigned long val=*p++-'0';   while (p<end) val=10*val+(*p++-'0');
      valp->val=val; code=INT;
    }
  else
    
    { switch(c)
      {      case '"': 
                    { valp->expression=make_string_denotation(scan_quoted_string());
                      code=STRING;
                    }
     
      break; case '(':
             case '{':
             case '[': ++nesting; input.push_prompt(c); code=c;
      break; case ')':
             case '}':
             case ']': --nesting; input.pop_prompt(); code=c;
      break; case '<':
             case '>':
             if (state==initial)
             { code= c=='<'
               ? input.shift()=='<' ? FORCEFROMFILE:  (input.unshift(),FROMFILE)
               : input.shift()=='>' ? ADDTOFILE : (input.unshift(),TOFILE) ;
    	   
    	   if ((skip_space(),c=input.shift())=='"')
    	     file_name=scan_quoted_string();
    	   else
    	   { file_name="";
    	       while (!std::isspace(c))
    	       { file_name.push_back(c); c=input.shift(); }
    	       input.unshift();
    	   }
                break;
             }
             prevent_termination=c;
             code = OPERATOR; valp->oper.priority=2;
             if (input.shift()=='=')
               valp->oper.id=id_table.match_literal(c=='<' ? "<=" : ">=");
             else
               { input.unshift();
                   valp->oper.id=id_table.match_literal(c=='<' ? "<" : ">");
               }
      break; case ':': prevent_termination=c;
        code = input.shift()=='=' ? BECOMES  : (input.unshift(),c);
      break; case '=':
             valp->oper.id = id_table.match_literal("=");
             valp->oper.priority = 2; // in case
    	 prevent_termination=code=c;
      break; case '!':
             if (input.shift()=='=')
             { code = OPERATOR;
               valp->oper.id = id_table.match_literal("!=");
               valp->oper.priority = 2;
               prevent_termination='=';
             }
             else
               code= (input.unshift(),c);
             // currently unused; might some day be factorial operator
      break; 
                 case '+': prevent_termination=code=c;
                    code = OPERATOR;
                    valp->oper.id = id_table.match_literal("+");
                    valp->oper.priority = 4;
             break; case '-': prevent_termination=c;
                    if (input.shift()=='>')
                       code= ARROW;
                    else
                    { input.unshift();
                      code = OPERATOR;
                      valp->oper.id = id_table.match_literal("-");
                      valp->oper.priority = 4;
                    }
             break; case '*': case '%': case '/': prevent_termination=c;
                    code = OPERATOR;
                    valp->oper.id =
                       id_table.match_literal(c=='*' ? "*" : c=='%' ? "%" : "/");
                    valp->oper.priority = 6;
             break; case '\\':
                    code = OPERATOR;
                    valp->oper.priority = 6;
                    if (input.shift()=='%')
                    { prevent_termination='%';
                      valp->oper.id = id_table.match_literal("\\%");
                    }
                    else
                    { input.unshift();
                      valp->oper.id = id_table.match_literal("\\");
                    }
             break; case '^': prevent_termination=c;
                    code = OPERATOR;
                    valp->oper.id = id_table.match_literal("^");
                    valp->oper.priority = 7; // exponentiation is right-associative
             break; case '#': prevent_termination=c;
                    code = OPERATOR;
                    valp->oper.id = id_table.match_literal("#");
                    valp->oper.priority = 8; // basic meaning is non-associative
             break;
             case '\n': state=ended; // and {\bf fall through}.
             default: code=c;
      }
    }
  input.locate(input.point(),locp->last_line,locp->last_column);
if (state==initial)
   state=normal;
  return code;
}


extern "C"
char* id_completion_func(const char* text, int state)
{ static size_t l; static id_type i,n;
  if (state==0)
  { i=0; n=main_hash_table->nr_entries();
    l=std::strlen(text); // fix length for during search
  }
  while (i<n)
    // |i| is initialised above when |state==0|, and incremented below
  { const char* s=main_hash_table->name_of(i++);
      // get stored identifier and increment loop
    if (std::strncmp(text,s,l) == 0) // is |text| a prefix of |s|?
      return strdup(s);
  }
  return NULL; /* if loop terminates, report failure */
}


Hash_table* main_hash_table=NULL;

char lexical_break_chars[] = " \t\n=<>+-*/\\%,;:()[]{}$@\"|";

Lexical_analyser* lex=NULL;

   }
}

