#define realex_version  "0.8.5" \
 
#include  "parsetree.h"
#include  "parser.tab.h"
#include  <iostream>
#include  <fstream>
#include  "buffer.h"
#include  "lexer.h"
#include  "version.h"
#include  <unistd.h> 
#include  "built-in-types.h"
#include  "constants.h"
#include  <stdexcept>
#include  "evaluator.h"
#include  <cstring>


#ifdef NREADLINE
#define readline nullptr
#define add_history nullptr
#define clear_history()
#else
#include <readline/readline.h>
#include <readline/history.h>
#endif

int yylex (YYSTYPE *, YYLTYPE *);
int yyparse( atlas::interpreter::expr_p* parsed_expr, int* verbosity );

namespace { 
            const char* keywords[] =
             {"quit"
             ,"set","let","in","begin","end"
             ,"if","then","else","elif","fi"
             ,"and","or","not"
             ,"while","do","od","next","for","from","downto"
             ,"true","false"
             ,"quiet","verbose"
             ,"whattype","showall","forget"
             ,nullptr};
 }

int yylex(YYSTYPE *valp, YYLTYPE *locp)
{ return atlas::interpreter::lex->get_token(valp,locp); }


void yyerror (YYLTYPE* locp, atlas::interpreter::expr_p* ,int* ,char const *s)
{ atlas::interpreter::main_input_buffer->show_range
  (std::cerr,
   locp->first_line, locp->first_column,
   locp->last_line,  locp->last_column);
  std::cerr << s << std::endl;
  atlas::interpreter::main_input_buffer->close_includes();
}

#ifndef NREADLINE
extern "C" char** do_completion(const char* text, int start, int end)
{
  if (start>0)
  { int i; char c; bool need_file=false;
    for (i=0; i<start; ++i)
      if (std::isspace(c=rl_line_buffer[i]))
        continue; // ignore space characters
      else if (c=='<' or c=='>')
        need_file=true;
      else
        break;

    if (need_file and i==start)
       // the text is preceded by one or more copies of \.<, \.>
      return nullptr; // signal that file name completion should be used
  }
  rl_attempted_completion_over = true;
    // don't try file name completion if we get here
  return rl_completion_matches(text,atlas::interpreter::id_completion_func);
}
#endif


int main(int argc, char** argv)
{ using namespace std; using namespace atlas::interpreter;

  
  bool use_readline = argc<2 or std::strcmp(argv[1],"-nr")!=0;
BufferedInput input_buffer(isatty(STDIN_FILENO) ? "expr> " : nullptr
                            ,use_readline ? readline : nullptr
			    ,use_readline ? add_history : nullptr);
  main_input_buffer= &input_buffer;
Hash_table hash; main_hash_table= &hash;
Lexical_analyser ana(input_buffer,hash,keywords,prim_names); lex=&ana;
Id_table main_table;  global_id_table=&main_table;
overload_table main_overload_table;
  global_overload_table=&main_overload_table;

  
  
  #ifndef NREADLINE
    using_history();
    rl_completer_word_break_characters = lexical_break_chars;
    rl_attempted_completion_function = do_completion; // set up input completion
  
  #endif
  ana.set_comment_delims('{','}');
  initialise_evaluator(); initialise_builtin_types();

  cout << "This is 'realex', version " realex_version " (compiled on " 
       << atlas::version::COMPILEDATE  <<
").\nIt is the programmable interpreter interface to the library (version " 
       << atlas::version::VERSION  << ") of\n"
       << atlas::version::NAME <<  ". http://www.liegroups.org/\n";

  last_value = shared_value (new tuple_value(0));
  last_type = void_type.copy();
   // |last_type| is a |type_ptr| defined in \.{evaluator.w}
  while (ana.reset()) // get a fresh line for lexical analyser, or quit
  { expr_p parse_tree;
    int old_verbosity=verbosity;
    ofstream redirect; // if opened, this will be closed at end of loop
    if (yyparse(&parse_tree,&verbosity)!=0)
      continue; // syntax error or non-expression
    if (verbosity!=0) // then some special action was requested
    { if (verbosity<0)
        break; // \.{quit} command
      if (verbosity==2 or verbosity==3)
        // indicates output redirection was requested
      { 
        { redirect.open(ana.scanned_file_name() ,ios_base::out |
             (verbosity==2 ? ios_base::trunc : ios_base::app));
          if (redirect.is_open())
            output_stream = &redirect;
          else
          { cerr << "Failed to open " << ana.scanned_file_name() << endl;
            continue;
          }
        }
        verbosity=old_verbosity; // verbosity change was temporary
      }
      if (verbosity==1) //
        cout << "Expression before type analysis: " << *parse_tree << endl;
    }
    
    { bool type_OK=false;
      try
      { expression_ptr e;
        last_type=analyse_types(*parse_tree,e); // move assignment of a |type_expr|
        type_OK=true;
        if (verbosity>0)
          cout << "Type found: " << last_type << endl 
    	   << "Converted expression: " << *e << endl;
        e->evaluate(expression_base::single_value);
        last_value=pop_value();
        static type_expr empty(empty_tuple());
        if (last_type!=empty)
          *output_stream << "Value: " << *last_value << endl;
        destroy_expr(parse_tree);
      }
      catch (runtime_error& err)
      { if (type_OK)
          cerr << "Runtime error:\n  " << err.what() << "\nEvaluation aborted.";
        else cerr << err.what();
        cerr << std::endl;
        reset_evaluator(); main_input_buffer->close_includes();
      }
      catch (logic_error& err)
      { cerr << "Internal error: " << err.what() << ", evaluation aborted.\n";
        reset_evaluator(); main_input_buffer->close_includes();
      }
      catch (exception& err)
      { cerr << err.what() << ", evaluation aborted.\n";
        reset_evaluator(); main_input_buffer->close_includes();
      }
    }
    output_stream= &cout; // reset output stream if it was changed
  }
  clear_history();
  // clean up (presumably disposes of the lines stored in history)
  cout << "Bye.\n";
  return 0;
}


