#include  "parsetree.h"
#include  "types.h" 
#include  "lexer.h"

namespace atlas
{ namespace interpreter
  {

void destroy_expr(expr_p p)  { delete p; }

expr_p make_int_denotation (int val)
{ return new expr(val); }

expr_p make_bool_denotation(bool val)
{ expr result (val ? 1 : 0);
  result.kind=boolean_denotation;
  return new expr(std::move(result));
}

expr_p make_string_denotation(std::string&& val)
 { return new expr(std::move(val)); }

expr mk_applied_identifier (id_type id)
{ expr result; result.kind=applied_identifier;
  result.identifier_variant=id; return result;
}
expr_p make_applied_identifier (id_type id)
 { return new expr(mk_applied_identifier(id)); }

expr_p make_dollar ()
{ expr_p result = new expr;
  result->kind=last_value_computed;
  return result;
}

raw_expr_list make_exprlist_node(expr_p e, raw_expr_list raw)
  { expr_ptr saf(e); expr_list l(raw);
    l.push_front(std::move(*e));
    return l.release();
  }

raw_expr_list reverse_expr_list(raw_expr_list raw)
{ expr_list l(raw); l.reverse(); return l.release(); }

expr_p wrap_tuple_display(raw_expr_list l)
{ return new expr(expr_list(l),true); }
expr_p wrap_list_display(raw_expr_list l)
{ return new expr(expr_list(l),false); }

void destroy_exprlist(raw_expr_list l)
{ (expr_list(l)); }

expr mk_application_node(expr&& f, expr_list&& args)
{ app a(new application_node { std::move(f), expr() });
  if (args.singleton())
    a->arg.set_from(args.front());
  else
    a->arg=expr(std::move(args),true); // make into an argument tuple
  return expr(std::move(a)); // move construct application expression
}
expr_p make_application_node(expr_p f, raw_expr_list args)
 { expr_ptr ff(f);
   return new expr(mk_application_node(std::move(*f),expr_list(args))); }

expr mk_binary_call(id_type name, expr&& x, expr&& y)
{ expr_list args;
  args.push_front(std::move(y));
  args.push_front(std::move(x));
  return mk_application_node(mk_applied_identifier(name),std::move(args));
}
expr_p make_binary_call(id_type name, expr_p x, expr_p y)
 {
   return new
     expr(mk_binary_call(name,std::move(*expr_ptr(x)),std::move(*expr_ptr(y))));
 }

expr mk_unary_call(id_type name, expr&& arg)
{
  return expr(app(
      new application_node { mk_applied_identifier(name), std::move(arg) }));
}
expr_p make_unary_call(id_type name, expr_p a)
 { return new expr(mk_unary_call(name,std::move(*expr_ptr(a)))); }

raw_form_stack start_formula (expr_p e, id_type op, int prio)
{ form_stack result;
  result.push_front ( {std::move(*expr_ptr(e)), op, prio } );
  return result.release();
}

raw_form_stack start_unary_formula (id_type op, int prio)
{ form_stack result;
  result.push_front ( { expr(), op, prio } ); // leave |left_subtree| empty
  return result.release();
}

raw_form_stack extend_formula (raw_form_stack pre, expr_p ep,id_type op, int prio)
{ expr e(std::move(*expr_ptr (ep))); form_stack s(pre);
  while (not s.empty() and s.front().prio>=prio+prio%2)
    
    { if (s.front().left_subtree.kind==no_expr)
        e = mk_unary_call(s.front().op,std::move(e));
          // apply initial unary operator
      else
        e = mk_binary_call(s.front().op
                         ,std::move(s.front().left_subtree)
                         ,std::move(e));
      s.pop_front();
    }
  s.push_front({std::move(e),op,prio});
  return s.release();
}

expr_p end_formula (raw_form_stack pre, expr_p ep)
{ expr e(std::move(*expr_ptr (ep))); form_stack s(pre);
  while (not s.empty())
    
    { if (s.front().left_subtree.kind==no_expr)
        e = mk_unary_call(s.front().op,std::move(e));
          // apply initial unary operator
      else
        e = mk_binary_call(s.front().op
                         ,std::move(s.front().left_subtree)
                         ,std::move(e));
      s.pop_front();
    }
  return new expr(std::move(e));
}

void destroy_formula(raw_form_stack s)
{ (form_stack(s)); }

raw_patlist make_pattern_node(raw_patlist prev,raw_id_pat& body)
{ patlist l(prev); l.push_front(id_pat(body)); return l.release(); }

void destroy_pattern(raw_patlist p)
{ (patlist(p)); }

void destroy_id_pat(const raw_id_pat& p)
{ if ((p.kind & 0x2)!=0)
    destroy_pattern(p.sublist);
}

raw_patlist reverse_patlist(raw_patlist raw)
{ patlist p(raw);
  p.reverse();
  return p.release();
}

raw_let_list make_let_node(raw_id_pat& pattern, expr_p val)
{ let_list result;
  result.push_front ( { id_pat(pattern), std::move(*expr_ptr(val)) } );
  return result.release();
}

raw_let_list append_let_node(raw_let_list prev, raw_let_list cur)
{ let_list result(prev);
  result.push_front(std::move(let_list(cur).front()));
  return result.release();
 }

expr mk_let_expr_node(let_list&& decls, expr& body)
{ id_pat pattern; expr val;
  if (decls.singleton()) // single declaration
  { pattern=std::move(decls.front().pattern);
    val=std::move(decls.front().val);
  }
  else
  { patlist patl;
    val=expr(expr_list(),true); // make a tuple expression
    for (auto it=decls.begin(); not decls.at_end(it); ++it)
      // zip open |decls|, reversing
    { patl.push_front(std::move(it->pattern));
      val.sublist.push_front(std::move(it->val));
    }
    pattern = id_pat(std::move(patl));
  }
  return expr(let(new let_expr_node
    { std::move(pattern), std::move(val),std::move(body) }));

}
expr_p make_let_expr_node(raw_let_list decls, expr_p body)
 { return new expr(mk_let_expr_node(let_list(decls),*expr_ptr(body)));
 }

void destroy_letlist(raw_let_list l)
{ (let_list(l)); }

void destroy_type(type_p t) { (type_ptr(t)); }
void destroy_type_list(raw_type_list t) { (type_list(t)); }
  

expr mk_lambda_node(patlist&& pat_l, type_list&& type_l, expr& body)
{ id_pat pattern; type_expr parameter_type;
  if (type_l.singleton())
{ pattern=std::move(pat_l.front());
    parameter_type = std::move(type_l.front());
  }
  else
{ pattern=id_pat(std::move(pat_l));
    parameter_type=type_expr(std::move(type_l)); // make tuple type
  }
  return expr(lambda(new lambda_node
      {std::move(pattern),std::move(parameter_type),std::move(body)}));
}
expr_p make_lambda_node(raw_patlist p, raw_type_list tl, expr_p body)
 {
   return new expr(mk_lambda_node(patlist(p),type_list(tl),*expr_ptr(body)));
 }

expr mk_conditional_node(expr& c, expr& t, expr& e)
{
  return expr (cond (new 
      conditional_node { std::move(c), std::move(t), std::move(e) }));
}
expr_p make_conditional_node(expr_p c, expr_p t, expr_p e)
{
   return new expr(mk_conditional_node(*expr_ptr(c),*expr_ptr(t),*expr_ptr(e)));
}

expr mk_while_node(expr& c, expr& b)
{
  return expr (w_loop(new while_node { std::move(c), std::move(b)})) ;
}
expr_p make_while_node(expr_p c, expr_p b)
{ return new expr(mk_while_node(*expr_ptr(c),*expr_ptr(b))); }

expr mk_for_node(id_pat&& id, expr& ip, expr& b)
{
  return expr(f_loop(new
    for_node { std::move(id), std::move(ip), std::move(b) }));
}
expr_p make_for_node(raw_id_pat& id, expr_p ip, expr_p b)
{
   return new expr(mk_for_node(id_pat(id),*expr_ptr(ip),*expr_ptr(b)));
}

expr mk_cfor_node(id_type id, expr& cnt, expr& bnd, bool up, expr& b)
{
  return expr (c_loop(new 
   cfor_node { id, std::move(cnt),std::move(bnd),up,std::move(b) }));
}
expr_p make_cfor_node(id_type id, expr_p cnt, expr_p bnd, bool up, expr_p b)
{
   return new
     expr(mk_cfor_node(id,*expr_ptr(cnt),*expr_ptr(bnd),up,*expr_ptr(b)));
}

expr mk_subscription_node(expr& a, expr& i)
{
  return expr(sub(new subscription_node {std::move(a), std::move(i) }));
}
expr_p make_subscription_node(expr_p a, expr_p i)
{ return new expr(mk_subscription_node(*expr_ptr(a),*expr_ptr(i))); }

expr mk_cast(type_expr& type, expr& exp)
{
  return expr(cast(new cast_node { std::move(type), std::move(exp) }));
}
expr_p make_cast(type_p type, expr_p exp)
{ return new expr(mk_cast(*type_ptr(type),*expr_ptr(exp))); }

expr mk_op_cast(id_type name,type_expr& type)
{
  return expr(op_cast(new op_cast_node { name, std::move(type) }));
}
expr_p make_op_cast(id_type name,type_p type)
{ return new expr(mk_op_cast(name,*type_ptr(type))); }

expr mk_assignment(id_type lhs, expr& rhs)
{
  return expr(assignment(new assignment_node { lhs, std::move(rhs) }));
}
expr_p make_assignment(id_type lhs, expr_p rhs)
{ return new expr(mk_assignment(lhs,*expr_ptr(rhs))); }

expr mk_comp_ass(expr& lhs, expr& rhs)
{ return expr(comp_assignment(new 
  comp_assignment_node
  { lhs.subscription_variant->array.identifier_variant
  , std::move(lhs.subscription_variant->index)
  , std::move(rhs)
  }));
}
expr_p make_comp_ass(expr_p lhs, expr_p rhs)
{ return new expr(mk_comp_ass(*expr_ptr(lhs),*expr_ptr(rhs))); }

expr mk_sequence(expr& first, expr& last, bool forward)
{
  return expr(sequence(new 
    sequence_node { std::move(first), std::move(last), forward } ));
}
expr_p make_sequence(expr_p first, expr_p last, bool forward)
{ return new expr(mk_sequence(*expr_ptr(first),*expr_ptr(last),forward)); }

id_type lookup_identifier(const char* name)
{ return main_hash_table->match_literal(name); }

void include_file(int skip_seen)
{ if (not main_input_buffer->push_file
          (lex->scanned_file_name(),skip_seen!=0))
    main_input_buffer->close_includes();
     // nested include failure aborts all includes
}


expr::~expr()
{
  if (kind!=no_expr)
  {
    switch (kind)
    { 
      case integer_denotation: case boolean_denotation: break;
      case string_denotation:
        str_denotation_variant.~basic_string<char>(); break;
      
      case applied_identifier:
      case last_value_computed: break;
      
        case tuple_display:
        case list_display: sublist.~expr_list();
      break;
      
      case function_call:
        call_variant.~app();
      break;
      
      case let_expr: let_variant.~let();
      break;
      
      case lambda_expr: lambda_variant.~lambda();
      break;
      
      case conditional_expr:
        if_variant.~cond();
      break;
      
      case while_expr:
        while_variant.~w_loop();
      break;
      case for_expr:
        for_variant.~f_loop();
      break;
      case cfor_expr:
        cfor_variant.~c_loop();
      break;
      
      case subscription:
        subscription_variant.~sub();
      break;
      
      case cast_expr:
        cast_variant.~cast();
      break;
      
      case op_cast_expr:
        op_cast_variant.~op_cast();
      break;
      
      case ass_stat:
        assign_variant.~assignment();
      break;
      
      case comp_ass_stat:
        comp_assign_variant.~comp_assignment();
      break;
      
      case seq_expr:
        sequence_variant.~sequence();
      break;
       case no_expr: {}
    }
    kind = no_expr;
  }
}

void expr::set_from (expr& other)
{ assert(kind==no_expr);
  switch (kind=other.kind)
    { 
        case integer_denotation:
        case boolean_denotation:
          int_denotation_variant = other.int_denotation_variant; break;
        case string_denotation:
          new (&str_denotation_variant)
          std::string(std::move(other.str_denotation_variant));
          other.str_denotation_variant.~basic_string<char>();
        break;
      
      case applied_identifier: identifier_variant=other.identifier_variant; break;
      case last_value_computed: break;
      
        case tuple_display:
        case list_display:
        new (&sublist) expr_list (std::move(other.sublist));
          other.sublist.~expr_list();
        break;
      
         case function_call:
        new (&call_variant) app(std::move(other.call_variant)); break;
      
       case let_expr:
         new (&let_variant) let(std::move(other.let_variant));
       break;
      
      case lambda_expr: new (&lambda_variant) lambda(std::move(other.lambda_variant));
      break;
      
      case conditional_expr:
        new (&if_variant) cond(std::move(other.if_variant));
      break;
      
        case while_expr:
          new(&while_variant) w_loop(std::move(other.while_variant));
        break;
        case for_expr:
          new(&for_variant) f_loop(std::move(other.for_variant));
        break;
        case cfor_expr:
           new(&cfor_variant) c_loop(std::move(other.cfor_variant));
        break;
      
      case subscription:
        new (&subscription_variant) sub(std::move(other.subscription_variant));
      break;
      
      case cast_expr: new (&cast_variant) cast (std::move(other.cast_variant));
      break;
      
      case op_cast_expr:
        new (&op_cast_variant) op_cast(std::move(other.op_cast_variant));
      break;
      
      case ass_stat:
        new (&assign_variant) assignment(std::move(other.assign_variant));
      break;
      
      case comp_ass_stat:
        new (&comp_assign_variant)
        comp_assignment(std::move(other.comp_assign_variant));
      break;
      
      case seq_expr:
        new (&sequence_variant) sequence(std::move(other.sequence_variant));
      break;
        case no_expr: {}
    }
  other.kind = no_expr;
}

expr::expr (expr&& other) : kind(no_expr) 
{ set_from(other); }

void expr::operator= (expr&& other)
{
  if (this!=&other)
  { this->~expr();
    set_from(other);
  }
}

std::ostream& operator<< (std::ostream& out, const expr& e)
{ switch (e.kind)
  { 
    case integer_denotation: out << e.int_denotation_variant; break;
    case boolean_denotation:
      out << (e.int_denotation_variant!=0 ? "true" : "false"); break;
    case string_denotation:
      out << '"' << e.str_denotation_variant << '"'; break;
    
    case applied_identifier:
      out << main_hash_table->name_of(e.identifier_variant);
    break;
    case last_value_computed: out << '$';  break;
    
    case tuple_display:
    { const expr_list& l=e.sublist;
      if (l.empty()) out << "()";
      else
      { for (auto it=l.begin(); not l.at_end(it); ++it)
          out << (it==l.begin() ? '(' : ',') << *it;
        out << ')';
      }
    }
    break;
    case list_display:
    { const expr_list& l=e.sublist;
      if (l.empty()) out << "[]";
      else
      { for (auto it=l.begin(); not l.at_end(it); ++it)
          out << (it==l.begin() ? '[' : ',') << *it;
        out << ']';
      }
    }
    break;
    
    case function_call:
    { const app& a=e.call_variant;
      const expr& fun=a->fun; const expr& arg=a->arg;
      if (fun.kind==applied_identifier) out << fun;
      else out << '(' << fun << ')';
      if (arg.kind==tuple_display) out << arg;
      else out << '(' << arg << ')';
    }
    break;
    
    case let_expr:
    { const let& lexp=e.let_variant;
      out << "let " << lexp->pattern << '=' << lexp->val <<	 " in " << lexp->body;
    }
    break;
    
    case lambda_expr:
    { const lambda& fun=e.lambda_variant;
      out << '(' << fun->parameter_type << ' ' << fun->pattern << "):" << fun->body;
    }
    break;
    
    case conditional_expr:
    { const cond& c=e.if_variant; out << " if " << c->condition 
      << " then " << c->then_branch << " else " << c->else_branch << " fi ";
    }
    break;
    
    case while_expr:
    { const w_loop& w=e.while_variant;
     out << " while " << w->condition << " do " << w->body << " od ";
    }
    break;
    case for_expr:
    { const f_loop& f=e.for_variant;
      const patlist& pl = f->id.sublist;
    const id_pat& index = pl.front();
      const id_pat& entry = *++pl.begin();
    out << " for " << entry;
      if (index.kind==0x1)
        out << '@' << index;
      out << " in " << f->in_part << " do " << f->body << " od ";
    }
    break;
    case cfor_expr:
    { const c_loop& c=e.cfor_variant;
    out << " for " << main_hash_table->name_of(c->id) << ": " << c->count;
      if (c->bound.kind!=tuple_display or c->bound.sublist.empty())
        out << (c->up!=0 ? " from " : " downto ") << c->bound;
      out << " do " << c->body << " od ";
    }
    break;
    
    case subscription:
    { const sub& s=e.subscription_variant;
      const expr& i=s->index;
      out << s->array << '[';
      if (i.kind!=tuple_display) out << i;
      else
        for (auto it=i.sublist.begin(); not i.sublist.at_end(it); ++it)
          out << (it==i.sublist.begin() ? "" : ",") << *it;
      out << ']';
    }
    break;
    
    case cast_expr:
    { const cast& c = e.cast_variant;
      out << c->type << ':' << c->exp ;
    }
    break;
    
    case op_cast_expr:
    { const op_cast& c = e.op_cast_variant;
      out << main_hash_table->name_of(c->oper) << '@' << c->type;
    }
    break;
    
    case ass_stat:
    { const assignment& ass = e.assign_variant;
      out << main_hash_table->name_of(ass->lhs) << ":=" << ass->rhs ;
    }
    break;
    
    case comp_ass_stat:
    { const comp_assignment& ass = e.comp_assign_variant;
      out << main_hash_table->name_of(ass->aggr) << '[' << ass->index << "]:="
          << ass->rhs ;
    }
    break;
    
    case seq_expr:
    { const sequence& seq = e.sequence_variant;
      out << seq->first << (seq->forward ? ";" : " next ") << seq->last ;
    }
    break;
     case no_expr: {}
  }
  return out;
}

bool is_empty(const expr& e)
{ return e.kind==tuple_display and e.sublist.empty(); }

std::ostream& operator<< (std::ostream& out, const id_pat& p)
{ if ((p.kind & 0x2)!=0)
  { out << '(';
    if (not p.sublist.empty())
    { auto it = p.sublist.begin();
      out << *it;
      for ( ++it ; not p.sublist.at_end(it); ++it)
        out << ',' << *it;
    }
    out << ')';
  }
  if (p.kind==0x3) // both parts present
    out << ':';
  if ((p.kind & 0x1)!=0)
    out << main_hash_table->name_of(p.name);
  return out;
}

  }
}

