
#ifndef PARSETREE_H
#define PARSETREE_H

#include <iostream>
#include <memory>
#include <string>

#include "buffer.h" 

#include "sl_list.h"

#include "types.h"

namespace atlas
{
  namespace interpreter
  {


struct expr;

typedef Hash_table::id_type id_type;

typedef containers::simple_list<expr> expr_list;
typedef containers::sl_node<expr>* raw_expr_list;
  

typedef std::unique_ptr<struct application_node> app;

typedef containers::simple_list<struct id_pat> patlist;
typedef containers::sl_node<struct id_pat>* raw_patlist;

struct raw_id_pat
{ id_type name;
  unsigned char kind;
  // bits 0,1: whether pattern has a name, respectively sublist
  raw_patlist sublist;
};
struct id_pat
{ id_type name;
  unsigned char kind;
  // bits 0,1: whether pattern has a name, respectively sublist
  patlist sublist;

  id_pat() :name(), kind(0x0), sublist() {}
  id_pat(const raw_id_pat& x)
   : name(x.name), kind(x.kind)
  , sublist((kind & 0x2)==0 ? nullptr : x.sublist)
  {}
  id_pat (id_type n, unsigned char k, patlist&& l)
  : name(n), kind(k), sublist(std::move(l)) {}
  id_pat(patlist&& l): name(), kind(0x2),  sublist(std::move(l)) {}

  id_pat (const id_pat& x) =  delete ;
id_pat (id_pat&& x) =  default ;
id_pat& operator=(const id_pat& x) =  delete ;
id_pat& operator=(id_pat&& x) =  default ;
  raw_id_pat release()
  { return { name, kind, sublist.release() }; }
};

typedef std::unique_ptr<struct let_expr_node> let;

typedef std::unique_ptr<struct lambda_node> lambda;

typedef std::unique_ptr<struct conditional_node> cond;

typedef std::unique_ptr<struct while_node> w_loop;
typedef std::unique_ptr<struct for_node> f_loop;
typedef std::unique_ptr<struct cfor_node> c_loop;

typedef std::unique_ptr<struct subscription_node> sub;

typedef std::unique_ptr<struct cast_node> cast;

typedef std::unique_ptr<struct op_cast_node> op_cast;

typedef std::unique_ptr<struct assignment_node> assignment;

typedef std::unique_ptr<struct comp_assignment_node> comp_assignment;

typedef std::unique_ptr<struct sequence_node> sequence;


enum expr_kind 
 { 
   integer_denotation, string_denotation, boolean_denotation, 
   
   applied_identifier,
   last_value_computed, 
   
    tuple_display, list_display, 
   
    function_call, 
   
    let_expr, 
   
   lambda_expr,
   
    conditional_expr, 
   
    while_expr, for_expr, cfor_expr, 
   
    subscription, 
   
   cast_expr, 
   
   op_cast_expr, 
   
   ass_stat, 
   
   comp_ass_stat, 
   
   seq_expr, 
 no_expr };
struct expr {
  expr_kind kind;
  union { 
          int int_denotation_variant;
          std::string str_denotation_variant;
          
          id_type identifier_variant;
          
          expr_list sublist;
          
          app call_variant;
          
          let let_variant;
          
          lambda lambda_variant;
          
          cond if_variant;
          
          w_loop while_variant;
          f_loop for_variant;
          c_loop cfor_variant;
          
          sub subscription_variant;
          
          cast cast_variant;
          
          op_cast op_cast_variant;
          
          assignment assign_variant;
          
          comp_assignment comp_assign_variant;
          
          sequence sequence_variant;
 };

  
  expr() : kind(no_expr){}
  ~expr(); 
  
  void set_from (expr& other);
  expr (expr&& other);
  void operator= (expr&& other);
  
    explicit expr(int n)
    : kind(integer_denotation), int_denotation_variant(n) {}
    expr(bool b) =  delete ;
    explicit expr(std::string&& s)
     : kind(string_denotation)
     , str_denotation_variant(std::move(s)) {}
  
  expr(expr_list&& nodes, bool is_tuple)
   : kind(is_tuple ? tuple_display : list_display)
   , sublist(std::move(nodes))
   {}
  
  explicit expr(app&& fx)
   : kind(function_call)
   , call_variant(std::move(fx))
   {}
  
  explicit expr(let&& declaration)
   : kind(let_expr)
   , let_variant(std::move(declaration))
   {}
  
  explicit expr(lambda&& fun)
   : kind(lambda_expr)
   , lambda_variant(std::move(fun))
   {}
  
  explicit expr(cond&& conditional)
   : kind(conditional_expr)
   , if_variant(std::move(conditional))
   {}
  
  explicit expr(w_loop&& loop)
   : kind(while_expr)
   , while_variant(std::move(loop))
   {}
  explicit expr(f_loop&& loop)
   : kind(for_expr)
   , for_variant(std::move(loop))
   {}
  explicit expr(c_loop&& loop)
   : kind(cfor_expr)
   , cfor_variant(std::move(loop))
   {}
  
  explicit expr(sub&& s)
   : kind(subscription)
   , subscription_variant(std::move(s))
   {}
  
  explicit expr(cast&& c)
   : kind(cast_expr)
   , cast_variant(std::move(c))
   {}
  
  explicit expr(op_cast&& c)
   : kind(op_cast_expr)
   , op_cast_variant(std::move(c))
   {}
  
  explicit expr(assignment&& a)
   : kind(ass_stat)
   , assign_variant(std::move(a))
   {}
  
  explicit expr(comp_assignment&& ca)
   : kind(comp_ass_stat)
   , comp_assign_variant(std::move(ca))
   {}
  
  explicit expr(sequence&& s)
   : kind(seq_expr)
   , sequence_variant(std::move(s))
   {}

};
typedef expr* expr_p; // raw pointer type for use on parser stack
typedef std::unique_ptr<expr> expr_ptr;


struct application_node
{ expr fun; expr arg;

  application_node(expr&& fun, expr&& arg)
: fun(std::move(fun)), arg(std::move(arg)) {}
  // backward compatibility for gcc 4.6
};

struct formula_node { expr left_subtree; id_type op; int prio; };

typedef containers::simple_list<formula_node> form_stack;
typedef containers::sl_node<formula_node>* raw_form_stack;

struct let_pair { id_pat pattern; expr val; };
typedef containers::simple_list<let_pair> let_list;
typedef containers::sl_node<let_pair>* raw_let_list;

struct let_expr_node
{ id_pat pattern; expr val; expr body;

  let_expr_node(id_pat&& pattern, expr&& val, expr&& body)
: pattern(std::move(pattern)), val(std::move(val)), body(std::move(body)){}
  // backward compatibility for gcc 4.6
};

struct lambda_node
{ id_pat pattern; type_expr parameter_type; expr body;

  lambda_node(id_pat&& pattern, type_expr&& type, expr&& body)
: pattern(std::move(pattern))
  , parameter_type(std::move(type)), body(std::move(body)){}
  // backward compatibility for gcc 4.6
};

struct conditional_node
{ expr condition; expr then_branch; expr else_branch;

  conditional_node(expr&& condition, expr&& then_branch, expr&& else_branch)
: condition(std::move(condition))
  , then_branch(std::move(then_branch))
  , else_branch(std::move(else_branch)){}
  // backward compatibility for gcc 4.6
};

struct while_node
{ expr condition; expr body;

  while_node(expr&& condition, expr&& body)
: condition(std::move(condition))
  , body(std::move(body)){}
  // backward compatibility for gcc 4.6
};
struct for_node
{ struct id_pat id; expr in_part; expr body;

  for_node(id_pat&& id, expr&& in_part, expr&& body)
: id(std::move(id))
  , in_part(std::move(in_part))
  , body(std::move(body)){}
  // backward compatibility for gcc 4.6
};
struct cfor_node
{ id_type id; expr count; expr bound; bool up; expr body;

  cfor_node(id_type id, expr&& count, expr&& bound, bool up, expr&& body)
: id(id)
  , count(std::move(count))
  , bound(std::move(bound))
  , up(up)
  , body(std::move(body)){}
  // backward compatibility for gcc 4.6
};

struct subscription_node
{ expr array; expr index;

  subscription_node(expr&& array, expr&& index)
: array(std::move(array))
  , index(std::move(index)){}
  // backward compatibility for gcc 4.6
};

struct cast_node
{ type_expr type; expr exp;

  cast_node(type_expr&& type, expr&& exp)
: type(std::move(type))
  , exp(std::move(exp)){}
  // backward compatibility for gcc 4.6
};

struct op_cast_node
{ id_type oper; type_expr type;

  op_cast_node(id_type oper,type_expr&& type)
: oper(oper)
  , type(std::move(type)){}
  // backward compatibility for gcc 4.6
};

struct assignment_node
{ id_type lhs; expr rhs;

  assignment_node(id_type lhs, expr&& rhs)
: lhs(lhs)
  , rhs(std::move(rhs)){}
  // backward compatibility for gcc 4.6
};

struct comp_assignment_node
{ id_type aggr; expr index; expr rhs;

  comp_assignment_node(id_type aggr, expr&& index, expr&& rhs)
: aggr(std::move(aggr))
  , index(std::move(index))
  , rhs(std::move(rhs)){}
  // backward compatibility for gcc 4.6
};

struct sequence_node
{ expr first; expr last; bool forward;

  sequence_node(expr&& first, expr&& last, bool forward)
: first(std::move(first))
  , last(std::move(last))
  , forward(forward){}
  // backward compatibility for gcc 4.6
};


void destroy_expr(expr_p e);

expr_p make_int_denotation (int val);
expr_p make_bool_denotation(bool val);
expr_p make_string_denotation(std::string&& val);

expr_p make_applied_identifier (id_type id);
expr_p make_dollar();

raw_expr_list make_exprlist_node(expr_p e, raw_expr_list l);
raw_expr_list reverse_expr_list(raw_expr_list l);
expr_p wrap_tuple_display(raw_expr_list l);
expr_p wrap_list_display(raw_expr_list l);

void destroy_exprlist(raw_expr_list l);

expr_p make_application_node(expr_p f, raw_expr_list args);

expr_p make_unary_call(id_type name, expr_p arg);
expr_p make_binary_call(id_type name, expr_p x, expr_p y);

raw_let_list make_let_node(raw_id_pat& pattern, expr_p val);
raw_let_list append_let_node(raw_let_list prev, raw_let_list cur);
expr_p make_let_expr_node(raw_let_list decls, expr_p body);

void destroy_letlist(raw_let_list l);

void destroy_type(type_p t);
void destroy_type_list(raw_type_list t);

expr_p make_lambda_node(raw_patlist pat_l, raw_type_list type_l, expr_p body);

expr_p make_conditional_node(expr_p c, expr_p t, expr_p e);

expr_p make_while_node(expr_p c, expr_p b);
expr_p make_for_node(raw_id_pat& id, expr_p ip, expr_p b);
expr_p make_cfor_node(id_type id, expr_p count, expr_p bound, bool up, expr_p b);

expr_p make_subscription_node(expr_p a, expr_p i);

expr_p make_cast(type_p type, expr_p exp);

expr_p make_op_cast(id_type name,type_p type);

expr_p make_assignment(id_type lhs, expr_p rhs);

expr_p make_comp_ass(expr_p lhs, expr_p rhs);

expr_p make_sequence(expr_p first, expr_p last, bool forward);

id_type lookup_identifier(const char*);
void include_file(int skip_seen);


raw_form_stack start_formula (expr_p e, id_type op, int prio);
raw_form_stack start_unary_formula (id_type op, int prio);
raw_form_stack extend_formula (raw_form_stack pre, expr_p e,id_type op, int prio);
expr_p end_formula (raw_form_stack pre, expr_p e);
void destroy_formula(raw_form_stack s);

raw_patlist make_pattern_node(raw_patlist prev,raw_id_pat& body);

void destroy_pattern(raw_patlist p);
void destroy_id_pat(const raw_id_pat& p);
raw_patlist reverse_patlist(raw_patlist p);



std::ostream& operator<< (std::ostream& out, const expr& e);

bool is_empty(const expr& e);

std::ostream& operator<< (std::ostream& out, const id_pat& p);

  }
}

#endif

